'''
Tests for the algo.multicorr module.
'''

import unittest
import numpy as np
import matplotlib.pyplot as plt
import os
import sys
from io import StringIO
import cv2

import ncempy.algo.multicorr as mc

class test_shiftedimages(unittest.TestCase):
    """Test the multicorr module on two shifted images"""

    def test_parse_input(self):
        '''
        Test to make sure inputs are properly accounted for.
        '''


        g1 = 1
        g2 = np.zeros([4,4])
        with self.assertRaises(TypeError):
            mc.parse_input(g1, g2) # both must be arrays
        with self.assertRaises(TypeError):
            mc.parse_input(g2, g1)

        # tests input sanitation.
        saved_stdout = sys.stdout
        try:
            out = StringIO()
            sys.stdout = out
            mc.parse_input(g2, g2, 'dummy', 4)
            output = out.getvalue().strip()
            self.assertEqual(output, 'Unknown method used, setting to cross')

            out = StringIO()
            sys.stdout = out
            mc.parse_input(g2, g2, 'cross', 23.1)
            output = out.getvalue().strip()
            self.assertEqual(output, 'Upsample factor is not an integer, rounding down')

            out = StringIO()
            sys.stdout = out
            mc.parse_input(g2, g2, 'cross', -23.1)
            output = out.getvalue().strip()
            self.assertEqual(output, 'Upsample factor is not an integer, rounding down\nUpsample factor is < 1, setting to 1')

            out = StringIO()
            sys.stdout = out
            mc.parse_input(g2, g2, 'cross', 'phase')
            output = out.getvalue().strip()
            self.assertEqual(output, 'Upsample factor is not an integer or float, setting to 1')
        finally:
            sys.stdout = saved_stdout

        with self.assertRaises(TypeError):
            mc.parse_input(np.zeros([2,3]), np.zeros([3,2]), 'phase', 1)

        with self.assertRaises(TypeError):
            mc.initial_correlation_image(g2, g2, 'tester', 1)

    def test_multicorr(self):
        '''
        Test to check if the correlation is working.
        '''
        g2 = np.zeros((3,3))
        test = mc.multicorr(np.fft.fft2(g2), np.fft.fft2(g2), 'phase', 3)
        print(test)
        assert list(test) == [0, 0]

        # filename = '/Users/Tom/Downloads/matt_beard.jpg'
        # filename_shifted = '/Users/Tom/Downloads/matt_beard_shifted.jpg'
        # G1 = cv2.imread(filename, 0)
        # G1 = G1[0:100, 0:100]
        G1 = np.zeros((100,100))
        G1[42,35] = 12
        # G2 = cv2.imread(filename_shifted, 0)
        G2 = np.real(np.fft.ifft2(mc.imageShifter(np.fft.fft2(G1), [-30, 10])))
        print(G1.shape, G2.shape)
        # plt.figure(0)
        # plt.imshow(np.concatenate((G1, G2)))
        # plt.show(block = True)
        imageCorr = mc.initial_correlation_image(np.fft.fft2(G1), np.fft.fft2(G2), 'phase', 2)

        # test to see if correlation image looks the same
        np.testing.assert_almost_equal(np.exp(1j * np.angle((np.multiply(np.fft.fft2(G2), np.conj(np.fft.fft2(G1)))))), imageCorr, decimal=4)

        # plt.figure(1)
        # plt.imshow(np.real(np.fft.ifft2(imageCorr)))
        # plt.show(block = True )
        out_phase = mc.multicorr(np.fft.fft2(G1), (np.fft.fft2(G2)), 'phase', 2)
        self.assertEqual(out_phase, [-30.0, 10.0])

        # test cross correlation
        out_cross = mc.multicorr(np.fft.fft2(G1), np.fft.fft2(G2), 'cross', 1)
        self.assertEqual(out_cross, [-30.0, 10.0])

        # test hybrid correlation and dft upsample
        out_hybrid = mc.multicorr(np.fft.fft2(G1), np.fft.fft2(G2), 'hybrid', 3)
        out_hybrid = list(out_hybrid)
        np.testing.assert_almost_equal(out_hybrid, [-30.0, 10.0], decimal = 6)

        # plt.imshow(np.subtract(G1, np.real(np.fft.ifft2(out.G2shift))))
        # plt.show(block = True)


    def test_different_shifts(self):
        '''
        Tests to check if mod is working correctly
        '''
        # filename = '/Users/Tom/Downloads/matt_beard.jpg'
        # G1 = cv2.imread(filename, 0)
        # G1 = G1[0:501, 0:501]
        G1 = np.zeros((101,101))
        G1[55,55] = 12
        # plt.imshow(G1)
        # plt.show(block = True)
        shifts = [[3., 1.], [-3., -1.], [-3., 1.], [3.,-1.]]
        shifts2 = [[10.3, 14.1], [-10.3, -14.1], [-10.3, 14.1], [10.3,-14.1]]
        for i in shifts:
            # print(i)
            G2 = np.real(np.fft.ifft2(mc.imageShifter(np.fft.fft2(G1), i)))
            # plt.imshow(np.concatenate((G1, G2)))
            # plt.show(block = True)
            with self.subTest(i = i):
                out_phase = mc.multicorr(np.fft.fft2(G1), (np.fft.fft2(G2)), 'phase', 3)
                np.testing.assert_almost_equal(out_phase, i, decimal = 6)
                out_cross = mc.multicorr(np.fft.fft2(G1), np.fft.fft2(G2), 'cross', 3)
                np.testing.assert_almost_equal(out_cross, i, decimal = 6)
                out_hybrid = mc.multicorr(np.fft.fft2(G1), np.fft.fft2(G2), 'hybrid', 3)
                out_hybrid = list(out_hybrid)
                np.testing.assert_almost_equal(out_hybrid, i, decimal = 6)
        for i in shifts2:
            # print(i)
            G2 = np.real(np.fft.ifft2(mc.imageShifter(np.fft.fft2(G1), i)))
            with self.subTest(i = i):

                out_phase = mc.multicorr(np.fft.fft2(G1), (np.fft.fft2(G2)), 'phase', 10)
                out_phase = list(out_phase)
                np.testing.assert_almost_equal(out_phase, i, decimal = 2)

                out_hybrid = mc.multicorr(np.fft.fft2(G1), np.fft.fft2(G2), 'hybrid', 10)
                out_hybrid = list(out_hybrid)
                np.testing.assert_almost_equal(out_hybrid, i, decimal = 2)

                out_cross = mc.multicorr(np.fft.fft2(G1), np.fft.fft2(G2), 'cross', 10)
                out_cross = list(out_cross)
                np.testing.assert_almost_equal(out_cross, i, decimal = 2)

if __name__ == "__main__":
    unittest.main()



'''
Tests for the algo.distortion module.
'''

import unittest
import numpy as np
import matplotlib.pyplot as plt
import os

import ncempy.io.emd
import ncempy.algo.local_max
import ncempy.algo.distortion

class test_ringdiff(unittest.TestCase):
    '''
    Test the distortion module on ring diffraction patterns.
    '''
   
    
    def test_distortion(self):
        '''
        Test the distortion fitting algorithms to be used on ring diffraction patterns.
        '''
        
        plt.close('all')
        show=True
        
        # get an image
        femd = ncempy.io.emd.fileEMD('ncempy/test/resources/Pt_SAED_D910mm_single/Pt_SAED_D910mm_single.emd')
        data, dims = femd.get_emdgroup(femd.list_emds[0])
        
        img = data[0,:,:]
        dims = dims[1:3]
        
        # and a list of maxima
        points = ncempy.algo.local_max.local_max(img, 10, 600)
        points = ncempy.algo.local_max.points_todim(points, dims)
        
        # parameters
        center_init = ncempy.algo.local_max.points_todim((984,1032), dims)
        
        
        ## filter_ring
        # wrong input
        with self.assertRaises(TypeError):
            nopoints = ncempy.algo.distortion.filter_ring(np.array([[0,1,2,3,4],[5,6,7,8,9]]), center_init, (6e9, 8e9))
        with self.assertRaises(TypeError):
            nopoints = ncempy.algo.distortion.filter_ring(42, center_init, (6e9, 8e9))    
        with self.assertRaises(TypeError):
            nopoints = ncempy.algo.distortion.filter_ring(points, (1,2,3,4), (6e9, 8e9))        
            
        # no points
        nopoints = ncempy.algo.distortion.filter_ring(points, center_init, (8e9, 6e9))
        self.assertIsNone(nopoints)
        
        # feed through center
        self.assertTrue(np.array_equal(center_init, ncempy.algo.distortion.filter_ring(center_init, center_init, (0,0))))
        
        # working
        points = ncempy.algo.distortion.filter_ring(points, (center_init[0,0], center_init[0,1]), (6e9, 8e9))
        self.assertIsNotNone(points)
        points = ncempy.algo.distortion.filter_ring(points, center_init, (6e9, 8e9))
        self.assertIsNotNone(points)
        
        
        ## points_topolar
        # wrong input
        with self.assertRaises(TypeError):
            nopoints_plr = ncempy.algo.distortion.points_topolar(42, center_init)
        with self.assertRaises(TypeError):
            nopoints_plr = ncempy.algo.distortion.points_topolar(points, (1,2,3))
        with self.assertRaises(TypeError):
            nopoints_plr = ncempy.algo.distortion.points_topolar(np.array([[0,1,2],[3,4,5],[6,7,8]]), center_init)
        
        # working
        points_plr = ncempy.algo.distortion.points_topolar(points, center_init)
        
        
        ## plot_ringpolar
        # wrong input
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.distortion.plot_ringpolar(np.array([0,1,2,3,4]), dims)
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.distortion.plot_ringpolar(42, dims)
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.distortion.plot_ringpolar(points_plr, dims[0])
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.distortion.plot_ringpolar(points_plr, 42)
                    
        # working
        plot = ncempy.algo.distortion.plot_ringpolar(points_plr, dims, show=show)
        
        
        ## optimize_center
        # wrong input
        with self.assertRaises(TypeError):
            nocenter = ncempy.algo.distortion.optimize_center(np.array([0,1,2,3,4]), center_init)
        with self.assertRaises(TypeError):
            nocenter = ncempy.algo.distortion.optimize_center(42, center_init)
        with self.assertRaises(TypeError):
            nocenter = ncempy.algo.distortion.optimize_center(points, (1,2,3,4,5))
        with self.assertRaises(TypeError):
            nocenter = ncempy.algo.distortion.optimize_center(points, 'string')
        
        # get warning
        #center = ncempy.algo.distortion.optimize_center(points, center_init, maxfev=1)
        # working
        center = ncempy.algo.distortion.optimize_center(points, center_init, verbose=True)
        
        points_plr = ncempy.algo.distortion.points_topolar(points, center)
        plot = ncempy.algo.distortion.plot_ringpolar(points_plr, dims, show=show)
    
        ## optimize_distortion
        # wrong input
        with self.assertRaises(TypeError):
            nodists = ncempy.algo.distortion.optimize_distortion(np.array([[0,1,2,3,4],[5,6,7,8]]), (2,3,4))
        with self.assertRaises(TypeError):
            nodists = ncempy.algo.distortion.optimize_distortion(42, (2,3,4))        
        with self.assertRaises(TypeError):
            nodists = ncempy.algo.distortion.optimize_distortion(points_plr, 2)
            
        # working
        dists = ncempy.algo.distortion.optimize_distortion(points_plr, (2,3,4))
        
        ## plot_distpolar
        # wrong input
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.distortion.plot_distpolar(np.array([[0,1,2,3],[4,5,6,7]]), dims, dists, (2,3,4))
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.distortion.plot_distpolar(42, dims, dists, (2,3,4))
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.distortion.plot_distpolar(points_plr, dims[0], dists, (2,3,4))
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.distortion.plot_distpolar(points_plr, dims, dists[0:3], (2,3,4))
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.distortion.plot_distpolar(points_plr, dims, dists, 42)
            
        
        # try a bunch of fits
        for ns in ( (2,), (2,3), (2,3,4), (2,3,4,5), (4,) ):
     
            dists = ncempy.algo.distortion.optimize_distortion(points_plr, ns, verbose=True)
            plot = ncempy.algo.distortion.plot_distpolar(points_plr, dims, dists, ns, show=show)
            

        # wait for plots
        if show:
            plt.show()
    

# to test with unittest runner
if __name__ == '__main__':
    unittest.main()



'''
This module provides an interface to the Berkeley EMD file format.

See https://emdatasets.com/ for more details.
'''

import numpy as np
import h5py
import datetime


class fileEMD:
    '''Class to represent EMD files. 
    
    Implemented for spec 0.2 using the recommended layout for metadata.
    
    Meant to provide convenience functions for commonly occuring tasks. This means that you will still want to acces fileEMD.file_hdl to manipulate the HDF5 file for not so commonly occuring tasks.
    
    Parameters:
        filename (str):    Name of the EMD file.
        readonly (bool):    Set to open in read only mode.
    
    Note:
        The EMD module does not currently have a simplified "emdReader" like MRC, SER and DM. This will be offered in a future update.
    
    Example:
        Open an Berkeley EMD file. List the available data sets. Load a 3D data set and plot the first image. 
        
            >>> from ncempy.io import emd
            >>> import matplotlib.pyplot as plt
            >>> emd1 = emd.fileEMD('filename.emd')
            >>> [print(dataGroup.name) for dataGroup in emd1.list_emds]; #Use the builtin list_emds variable to print all available EMD datasets
            >>> data1,dims1 = emd1.get_emdgroup(emd1.list_emds[0]) #load the first full data array and dimension information
            >>> fg1,ax1 = plt.subplots(1,1)
            >>> ax1.imshow(data1[0,:,:],extent=(dims1[1][0][0],dims1[1][0][-1],dims1[2][0][0],dims1[2][0][-1])) #the extent uses the first and last array values of hte dimension vectors
            >>> ax1.set(xlabel='{0[1]} ({0[2]})'.format(dims1[1]),ylabel='{0[1]} ({0[2]})'.format(dims1[2])) #label the axes with the name and units of each dimension vector
            >>> plt.show()
            >>> del emd1 #close the emd file
    '''
    
    def __init__(self, filename, readonly=False):
        '''Init opening/creating the file.
        
        '''
        
        ## necessary declarations in case something goes bad
        self.file_hdl = None
        
        # convenience handles to access the data in the emd file, everything can as well be accessed using the file_hdl
        self.version = None
        self.data = None
        self.microscope = None
        self.sample = None
        self.user = None
        self.comments = None
        self.list_emds = []             # list of HDF5 groups with emd_data_type type
        
        # check for string
        if not isinstance(filename, str):
            raise TypeError('Filename is supposed to be a string!')

        # try opening the file
        if readonly:
            try:
                self.file_hdl = h5py.File(filename, 'r')
            except:
                print('Error opening file for readonly: "{}"'.format(filename))
                raise
        else:
            try:
                self.file_hdl = h5py.File(filename, 'a')
            except:
                print('Error opening file for read/write: "{}"'.format(filename))
                raise
        
        
        # if we got a working file
        if self.file_hdl:        
        
            # check version information
            if 'version_major' in self.file_hdl.attrs and 'version_minor' in self.file_hdl.attrs:
                # read version information
                self.version = (self.file_hdl.attrs['version_major'], self.file_hdl.attrs['version_minor'])
                # compare to implementation
                if not self.version == (0,2):
                    print('WARNING: You are reading a version {}.{} EMD file, this implementation assumes version 0.2!'.format(self.version[0], self.version[1]))
            else:
                # set version information
                if not readonly:
                    self.file_hdl.attrs['version_major'] = 0
                    self.file_hdl.attrs['version_minor'] = 2
                
            # check for data group
            if not 'data' in self.file_hdl:
                if not readonly:
                    self.data = self.file_hdl.create_group('data')
            else:
                self.data = self.file_hdl['data']
                
            # check for data group
            if not 'microscope' in self.file_hdl:
                if not readonly:
                    self.microscope = self.file_hdl.create_group('microscope')
            else:
                self.microscope = self.file_hdl['microscope']
                
            # check for data group
            if not 'sample' in self.file_hdl:
                if not readonly:
                    self.sample = self.file_hdl.create_group('sample')
            else:
                self.sample = self.file_hdl['sample']
                
            # check for data group
            if not 'user' in self.file_hdl:
                if not readonly:
                    self.user = self.file_hdl.create_group('user')
            else:
                self.user = self.file_hdl['user']
                
            # check for data group
            if not 'comments' in self.file_hdl:
                if not readonly:
                    self.comments = self.file_hdl.create_group('comments')
            else:
                self.comments = self.file_hdl['comments']
                
            # find emd_data_type groups in the file
            self.list_emds = self.find_emdgroups(self.file_hdl)
            

    def __del__(self):
        '''Destructor for EMD file object. 
        
        '''
        # close the file
        #if(not self.file_hdl.closed):
        self.file_hdl.close()

    def __enter__(self):
        '''Implement python's with staement
        
        '''
        return self
        
    def __exit__(self,type,value,traceback):
        '''Implement python's with statment
        and close the file via __del__()
        '''
        self.__del__()
        return None
        
    def find_emdgroups(self, parent):
        '''Find all emd_data_type groups within the group parent and return a list of references to their HDF5 groups.
        
        Parameters:
            parent (h5py._hl.group.Group):  Handle to the parent group.
            
        Returns:
            (list):    A list of h5py._hl.group.Group handles to children groups being emd_data_type groups.
            
        '''
        
        emds = []
        
        # recursive function to run and retrieve groups with emd_group_type set to 1
        def proc_group(group, emds):
            # take a look at each item in the group
            for item in group:
                # check if group
                if group.get(item, getclass=True) == h5py._hl.group.Group:
                    item = group.get(item)
                    # check if emd_group_type
                    if 'emd_group_type' in item.attrs:
                        if item.attrs['emd_group_type'] == 1:
                            emds.append(item)
                    # process subgroups
                    proc_group(item, emds)
        
        # run
        proc_group(parent, emds)
        
        return emds


    def get_emdgroup(self, group):
        '''Get the emdtype data saved in in group.
        
        Parameters:
            group (h5py._hl.group.Group):   Reference to the emdtype HDF5 group.
        
        Returns:
            (tuple/None):    None or tuple containing:
            
                np.ndarray:    The data of the emdtype group.
                
                list:    List of dimension vectors plus labels and units.
                
        '''
        
        # check input
        if not isinstance(group, h5py._hl.group.Group):
            raise TypeError('group needs to refer to a valid HDF5 group!')
            
        if not 'emd_group_type' in group.attrs:
            raise TypeError('group is not a emd_group_type group!')
        if not group.attrs['emd_group_type'] == 1:
            raise TypeError('group is not a emd_group_type group!')

        # retrieve data
        try:
            # get the data
            data = group['data'][:]
            
            # get the dims
            dims = []
            for i in range(len(data.shape)):
                dim = group['dim{}'.format(i+1)]
                # save them as (vector, name, units)
                
                if isinstance(dim.attrs['name'], np.ndarray):
                    name = dim.attrs['name'][0]
                else:
                    name = dim.attrs['name']
                
                if isinstance(dim.attrs['units'], np.ndarray):
                    units = dim.attrs['units'][0]
                else:
                    units = dim.attrs['units']
                    
                dims.append( (dim[:], name.decode('utf-8'), units.decode('utf-8')) )
            
            dims = tuple(dims)
            
            return data, dims
            
        except:
            # if something goes wrong, return None
            print('Content of "{}" does not seem to be in emd specified shape'.format(group.name))
            
            return None
            

    def write_dim(self, label, dim, parent):
        '''Auxiliary function to write a dim dataset to parent.
        
        Input is not checked for sanity, so handle exceptions in call.
        
        Parameters:
            label (str):    Label for dataset, usually dim1, dim2, dimN.
            dim (tuple):    Tuple containing (data, name, units).
            parent (h5py._hl.group.Group):    HDF5 handle to parent group.
        
        Returns:
            (h5py._hl.group.Group):    HDF5 dataset handle referencing this dim.
            
        '''
        
        try:
            dset = parent.create_dataset(label, data=dim[0])
            dset.attrs['name'] = np.string_(dim[1])
            dset.attrs['units'] = np.string_(dim[2])
        except:
            raise RuntimeError('Error during writing dim dataset')
        
        return dset
        
        
    def put_emdgroup(self, label, data, dims, parent=None, overwrite=False):
        '''Put an emdtype dataset into the EMD file.
        
        Parameters:
            label (str):    Label for the emdtype group containing the dataset.
            data (np.ndarray):    Numpy array containing the data.
            dims (tuple):    Tuple containing the necessary dims as ((vec, name, units), (vec, name, units), ...)
            parent (h5py._hl.group.Group/None):    Parent for the emdtype group, if None it will be written to /data.
            overwrite (bool):    Set to force overwriting entry in EMD file.
        
        Returns:
            (h5py._hl.group.Group/None):    Group referencing this emdtype dataset or None if failed.
            
        '''
        
        # check input
        if not isinstance(label, str):
            raise TypeError('label needs to be string!')
        
        if not isinstance(data, np.ndarray):
            raise TypeError('data needs to be a numpy.ndarray!')
            
        try:
            assert len(dims) == len(data.shape)
            for i in range(len(dims)):
                assert len(dims[i]) == 3
                assert dims[i][0].shape[0] == data.shape[i]
        except:
            raise TypeError('Something wrong with the provided dims')
        
        # write stuff to HDF5
        
        # create group
        try:
            if parent:
                if label in parent:
                    if overwrite:
                        print('overwriting "{}" in "{}"'.format(label, parent.name))
                        del parent[label]
                    else:
                        print('"{}" already exists in "{}"'.format(label, parent.name))
                        raise RuntimeError('"{}" already exists in "{}"'.format(label, parent.name))
                grp = parent.create_group(label)
                
            else:
                if label in self.data:
                    if overwrite:
                        print('overwriting "{}" in "{}"'.format(label, self.data.name))
                        del self.data[label]
                    else:
                        print('"{}" already exists in "{}"'.format(label, self.data.name))
                        raise RuntimeError('"{}" already exists in "{}"'.format(label, self.data.name))

                grp = self.data.create_group(label)
                
            # add attribute
            grp.attrs['emd_group_type'] = 1
               
            # create dataset
            dset = grp.create_dataset('data', data=data)
             
            # create dim datasets
            for i in range(len(dims)):
                self.write_dim('dim{}'.format(i+1), dims[i], grp)
                    
            # update emds list
            self.list_emds = self.find_emdgroups(self.file_hdl)
                    
            return grp
           
        except:
            print('Something went wrong trying to write the dataset.')
                
            return None


    def put_comment(self, msg, timestamp=None):
        '''Create a comment in the EMD file.
        
        If timestamp already exists, the msg is appended to existing comment.
        
        Parameters:
            msg (str):    String of the message to save.
            timestamp (str/None):    Timestamp used as the key, defaults to the current UTC time.
        
        '''
        
        # check input
        if not isinstance(msg, str):
            raise TypeError('msg needs to be a string!')
        
        # create timestamp if missing
        if not timestamp:
            timestamp = datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S (UTC)')
        else:
            # try to convert given timestamp to string
            try:
                timestamp = str(timestamp)
            except:
                raise
        
        # write comment
        if timestamp in self.comments.attrs:
            # append to existing
            self.comments.attrs[timestamp] += np.string_('\n'+msg)
        
        else:
            # create new entry
            self.comments.attrs[timestamp] = np.string_(msg)
        

def defaultDims(data):
    ''' A helper function that can generate properly setup dim tuple
    with default values to allow quick writing of EMD files without
    the need to create these dim vectors.
    
    Parameters:
        data (ndarray) : The data that will be written to the EMD file. This is used to get the number of dims and their shape
        
    Returns:
        (tuple) : a properly formatter tuple of dim vectors as input to emd.put_emdgroup()
    '''
    num = data.ndim
    
    dims = []
    for ii in range(num):
        curDim = (np.linspace(0,data.shape[ii]-1,data.shape[ii]),'dim{}'.format(ii),'unit{}'.format(ii))
        dims.append(curDim)
    
    return dims


'''
A module to read MRC files in python and numpy.
Written according to MRC specification at http://bio3d.colorado.edu/imod/betaDoc/mrc_format.txt
ALso works with FEI MRC files which include a special header block with experimental information.
written by: Peter Ercius, percius@lbl.gov
'''

import numpy as np

class fileMRC:
    '''Init opening the file and reading in the header.
    Read in the data in MRC format and other useful information.
            
    Parameters:
        filename (str): string pointing to the filesystem location of the file.
        verbose (bool): if True, debug information is printed.
    Returns:
        (dict): A dictionary with keys data, voxelSize, filename, axisOrientations, {FEIinfo}
    
    Note:
        Most users will prefer to use the mrc.mrcReader() function to simply read
        the entire data set into memory with a single command.
    
    Example:
        Simply read in all data from disk into memory. This assumes the dataset is 3 dimensional:
        
            >>> from ncempy.io import mrc
            >>> import matplotlib.pyplot as plt
            >>> mrc1 = mrc.fileMRC('filename.mrc')
            >>> data = mrc.getDataset() #load the full data set and meta data
            >>> singleImage = mrc.getSlice(0) #get only 1 image from disk
    '''    
    def __init__(self, filename, verbose = False):

        # check for string
        if not isinstance(filename, str):
            raise TypeError('Filename is supposed to be a string')
            
        self.filename = filename

        # necessary declarations, if something fails
        self.fid = None
        
        self.dataOut = {} #will hold the data and metadata to output to the user after getDataset() call
        
        #Add a top level variable to indicate verbosee output for debugging
        self.v = verbose
        
        #Open the file and quit if the file does not exist
        try:
            self.fid = open(self.filename,'rb')
        except IOError as e:
            print("I/O error({0}): {1}".format(e.errno, e.strerror))
        
        #Store the original filename
        self.dataOut['filename'] = self.filename
        
        self.parseHeader()
        
        return None
    
    def __del__(self):
        '''Close the file.
        
        '''
        if(not self.fid.closed):
            if self.v:
                print('Closing input file: {}'.format(self.filename))
            self.fid.close()
        return None
    
    def __enter__(self):
        '''Implement python's with statement
        
        '''
        return self
        
    def __exit__(self,type,value,traceback):
        '''Implement python's with statement.
        Close the file using __del__
        
        '''
        self.__del__()
        return None
            
    def parseHeader(self):
        '''Read the header information which includes data type, data size, data
        shape, and metadata.
        
        Note:
            This header uses Fortran-style ordering. Numpy uses C-style ordering.
            The header is read in and then some attributes are reversed [::-1] at
            the end for output to the user to match C-ordering in numpy.
        
        '''
        
        #Always start at the beginnging of the file.
        self.fid.seek(0)
        
        #Read in the initial header values
        head1 = np.fromfile(self.fid,dtype=np.int32,count=10)
        if self.v:
            print('header1 = {}'.format(head1))
        #Set the number of pixels for each dimension
        self.dataSize = head1[0:3]
        if self.v:
            print('dataSize (fortran ordering) = {}'.format(self.dataSize))
        
        #Set the data type and convert to numpy type
        self.mrcType = head1[3]
        self.dataType = self._getMRCType(self.mrcType)
        if self.v:
            print('dataType = {}'.format(self.dataType))
        
        #Get the grid size
        self.gridSize = head1[7:10]
        if self.v:
            print('mrc defined gridSize = {}'.format(self.gridSize))
            
        #Get the physical volume size (always in Angstroms) (starting at byte #11 in the file).
        head2 = np.fromfile(self.fid,dtype=np.float32,count=6)
        
        self.volumeSize = head2[0:3]
        if self.v:
            print('mrc defined volumeSize = {}'.format(self.volumeSize))
        
        #calculate the voxel size based on volume and grid sizes
        if self.volumeSize.any() ==0 or self.gridSize.any() ==0:
            if self.v:
                print('Detected 0 volume or grid size. Setting voxel size to 1 (Ang).')
            self.voxelSize = np.ones(3) #use 1 as a voxel size if its not set in the file
        else:
            self.voxelSize = (self.volumeSize / np.float32(self.gridSize))
            if self.v:
                print('voxelSize (Ang) = {}'.format(self.voxelSize))

        #Pixel (cell) angles
        self.cellAngles = head2[3:6]
        if self.v:
            print('cellAngles = {}'.format(self.cellAngles))
            
        #Axis orientations. Tells which axes are X,Y,Z
        self.axisOrientations = np.fromfile(self.fid,dtype=np.int32,count=3)
        if self.v:
            print('axisOrientations = {}'.format(self.axisOrientations))
        
        self.Shape = [self.dataSize[x-1] for x in self.axisOrientations]
        if self.v:
            print('data shape = {}'.format(self.Shape))
        
        #Min, max,mean
        self.minMaxMean = np.fromfile(self.fid,dtype=np.int32,count=3)
        
        #Extra information (for FEI MRC file, extra(1) is the size of the FEI information encoded with the file in terms of 4 byte floats)
        self.extra = np.fromfile(self.fid,dtype=np.int32,count=34)
        
        #Numpy uses C-style ordering. The header is written in Fortran-Style ordering. Flip the order of everything useful
        if self.v:
            print('Note: The MRC header is written in Fortran-Style ordering, but Numpy uses C-style ordering. This program will now reverse the order (using [::-1]) of useful metadata: (dataSize, gridSize,volumeSize,voxelSize,cellAngles,axisOrientations)')
        self.dataSize = self.dataSize[::-1]
        self.gridSize = self.gridSize[::-1]
        self.volumeSize = self.volumeSize[::-1]
        self.voxelSize = self.voxelSize[::-1]
        self.cellAngles = self.cellAngles[::-1]
        self.axisOrientations = self.axisOrientations[::-1]
        self.Shape = self.Shape[::-1]
        
        #Move to the end of the normal header
        self.fid.seek(1024)
        
        #Read in the extended header if it exists (for FEI MRC files)
        if self.extra[1] != 0:
            pos1 = self.fid.tell()
            if self.v:
                print('Extra header found. Most likely and FEI-style MRC file.')
                print('Position before reading extra header: ' + str(pos1))
                print('Extra header size = ' +str(self.extra[1]))
            
            '''Read the extra FEI header described as follows:
             1 a_tilt  first Alpha tilt (deg)
             2 b_tilt  first Beta tilt (deg)
             3 x_stage  Stage x position (Unit=m. But if value>1, unit=???m)
             4 y_stage  Stage y position (Unit=m. But if value>1, unit=???m)
             5 z_stage  Stage z position (Unit=m. But if value>1, unit=???m)
             6 x_shift  Image shift x (Unit=m. But if value>1, unit=???m)
             7 y_shift  Image shift y (Unit=m. But if value>1, unit=???m)
             8 defocus  starting Defocus Unit=m. But if value>1, unit=???m)
             9 exp_time Exposure time (s)
             10 mean_int Mean value of image
             11 tilt_axis   Tilt axis (deg)
             12 pixel_size  Pixel size of image (m)
             13 magnification   Magnification used
             14 voltage accelerating voltage
             15 ??
            '''
            FEIinfoValues = np.fromfile(self.fid,dtype=np.float32,count=15)
            self.FEIinfo = {'a_tilt':FEIinfoValues[0],'b_tilt':FEIinfoValues[1],'x_stage':FEIinfoValues[2],'y_stage':FEIinfoValues[3],'z_stage':FEIinfoValues[4],'x_shift':FEIinfoValues[5],'y_shift':FEIinfoValues[6],'defocus':FEIinfoValues[7],'exposure_time':FEIinfoValues[8],'mean':FEIinfoValues[9],'tilt_axis':FEIinfoValues[10],'pixel_size':FEIinfoValues[11],'magnification':FEIinfoValues[12],'voltage':FEIinfoValues[13],'unknown':FEIinfoValues[14]}
            
            self.voxelSize[0] = 1. #set this to 1 but it should be the tilt angles. These can be non-uniform though.
            self.voxelSize[1] = self.FEIinfo['pixel_size']*1e10 #convert [m] to Angstroms as is the standard for MRCs
            self.voxelSize[2] = self.FEIinfo['pixel_size']*1e10
            
            if self.v:
                print('Extended header data')
                for aa,bb in self.FEIinfo.items():
                    print('{} = {}'.format(aa,bb))
            
        self.dataOffset = 1024+self.extra[1] #offset of the data from the start of the file
        
        #Add relevant information (metadata) to the output dictionary
        self.dataOut = {'voxelSize':self.voxelSize,'axisOrientations':self.axisOrientations,'cellAngles':self.cellAngles,'axisOrientations':self.axisOrientations,'filename':self.filename}
        if self.extra[1] != 0:
            self.dataOut['FEIinfo'] = self.FEIinfo
        
        return 1
    
    def getDataset(self):
        '''Read in the full data block and reshape to a matrix
        with C-style ordering.
        
        '''
        self.fid.seek(self.dataOffset,0) #move to the start of the data from the start of the file
        try:
            data1 = np.fromfile(self.fid,dtype=self.dataType,count=np.prod(self.dataSize))
            self.dataOut['data'] = data1.reshape(self.Shape)
        except MemoryError:
            print("Not enough memory to read in the full data set")        
        return self.dataOut
    
    def getSlice(self,num):
        '''Read in a slice of an MRC file. Useful for parsing through a large file without reading
        the entire data set into memory.
        
        Paremeters:
            num (int): Get the requested image.
        
        Returns:
            (ndarray): A 2D slice or a 3D set of slices along the first index
        
        Raises:
            IndexError: If num > the number of slices.
        '''
        # Check num is within the data array size bounds
        if num > (self.dataSize[0]-1):
            raise IndexError('Index {} is out of bounds for array with size {}'.format(num, self.dataSize[0]))
        
        self.fid.seek(self.dataOffset,0) #move to the start of the data by skipping the header
        imSize = self.dataSize[1]*self.dataSize[2] #size of each image in pixels
        byteSize = int(np.dtype(self.dataType).itemsize*imSize) #make sure this is an int32
        self.fid.seek(num*byteSize,1) #skip to the slice requested from the start of the data
        
        data1 = np.fromfile(self.fid,dtype=self.dataType,count=imSize) #read in the requested image
        data1 = data1.reshape((self.Shape[1],self.Shape[2])) #reshape the image
            
        return data1
    
    def _applyAxisOrientations(self,arrayIn):
        ''' This is untested and unused.
        
        '''
        return [arrayIn[x-1] for x in self.axisOrientations]
    
    def _getMRCType(self, dataType):
        '''Return the correct data type according to the official MRC type list:
        0 image : signed 8-bit bytes range -128 to 127
        1 image : 16-bit halfwords
        2 image : 32-bit reals
        3 transform : complex 16-bit integers
        4 transform : complex 32-bit reals
        6 image : unsigned 16-bit range 0 to 65535
        
        Parameters:
            dataType (int): The data type value encoded in an MRC header
            
        Returns:
            {numpy dtype}: The corresponding numpy data type.
        
        '''
        if dataType == 0:
            Type = np.int8
        elif dataType == 1:
            Type = np.int16
        elif dataType == 2:
            Type = np.float32
        elif dataType ==  6:
            Type = np.uint16
        else:
            print("Unsupported data type" + str(dataType)) #complex data types are currently unsupported
        return Type
#end class fileMRC

def mrcReader(fname,verbose=False):
    '''A simple function to read open a MRC, parse the header, and read the full
    data set.
    
    Parameters:
        fname (str): The name of the file to load
        
    Keywords:
        verbose (bool): Enable printing debug messages as the header is parsed.
        
    Returns:
        (dict): A dictionary containing the data and interesting metadata. The data
        is attached to the 'data' key.
        
    Example:
        Simply read in all data from disk into memory. This assumes the dataset is 3 dimensional:
            
            >>> from ncempy.io import mrc
            >>> import matplotlib.pyplot as plt
            >>> mrc1 = mrc.mrcReader('filename.mrc')
            >>> plt.imshow(mrc1['data'][0,:,:]) #show the first image in the data set
    '''
    with fileMRC(fname,verbose) as f1: #open the file and init the class
        im1 = f1.getDataset() #read in the dataset
    return im1 #return the data and metadata as a dictionary

def mrc2raw(fname):
    '''Writes the image data in an MRC file as binary file with the same file
    name and .raw ending. Data type and size are written in the file name.
    No other header information is retained.
    
    Parameters:
        fname (str): The name of the file to convert
    
    '''
    tomo = mrcReader(fname)
    rawName = tomo['filename'].rsplit('.',1)[0] + '_' + str(tomo['data'].dtype) + '_' + str(tomo['data'].shape) + '.raw'
    fid = open(rawName,'wb')
    fid.write(tomo['data']) #write out as C ordered data
    fid.close()
    
def mrc2emd(fname):
    '''Write an MRC file as an HDF5 file in EMD format with same file name and .emd ending.
    Header information is retained as attributes.
    
    TODO: Update this to use ncempy.emd class
    '''
    import h5py
    
    #Read in the MRC data and reshape to C-style ordering
    tomo = mrcReader(fname)
    
    #create the HDF5 file
    try:
        f1 = h5py.File(fname.rsplit('.mrc',1)[0] + '.emd','w') #w- will error if the file exists
    except:
        print("Problem opening file. Maybe it already exists?")
        f1.close()
        del tomo
        return 0

    #Create the axis vectors in nanometers. Standard MRC pixel size is in Angstroms
    xFull = np.linspace(0,tomo['voxelSize'][0]*tomo['data'].shape[0]-1,tomo['data'].shape[0]) 
    yFull = np.linspace(0,tomo['voxelSize'][1]*tomo['data'].shape[1]-1,tomo['data'].shape[1])
    zFull = np.linspace(0,tomo['voxelSize'][2]*tomo['data'].shape[2]-1,tomo['data'].shape[2])
    
    #Root data group
    dataTop = f1.create_group('data')

    #Create tilt series group
    tiltseriesGroup = dataTop.create_group('data')
    tiltseriesGroup.attrs['emd_group_type'] = np.int8(1)
    
    #Save the data to the EMD file and reshape it to a C-style array
    try:
        tiltDset = tiltseriesGroup.create_dataset('data',data=tomo['data'],compression='gzip',shuffle=True)
    except MemoryError:
        print("Not enough memory to write out data to EMD file")
        del tomo
        f1.close()
        return 0
        
    dim1 = tiltseriesGroup.create_dataset('dim1',data=xFull)
    dim1.attrs['name'] = np.string_('x')
    dim1.attrs['units'] = np.string_('')
    dim2 = tiltseriesGroup.create_dataset('dim2',data=yFull)
    dim2.attrs['name'] = np.string_('y')
    dim2.attrs['units'] = np.string_('')
    dim3 = tiltseriesGroup.create_dataset('dim3',data=zFull)
    dim3.attrs['name'] = np.string_('z')
    dim3.attrs['units'] = np.string_('')
    
    #Create the other groups
    scopeGroup = f1.create_group('Microscope')
    scopeGroup.attrs['voxel sizes'] = tomo['voxelSize']
    userGroup = f1.create_group('User')
    commentGroup = f1.create_group('Comments')
    
    #Possible way using keyword arguments to populate these fields
    #def greet_me(**kwargs):
    #if kwargs is not None:
    #    for key, value in kwargs.iteritems():
    #        print("%s == %s" %(key,value))
    
    f1.close()
    
    return 1
    
def mrcWriter(filename,data,pixelSize,forceWrite=False):
    '''Write out a MRC type file according to the specification at http://bio3d.colorado.edu/imod/doc/mrc_format.txt
    
    Parameters:
        filename (str): The name of the MRC file.
        data (ndarray): The array data to write to disk.
        pixelSize (tuple): The size of the pixel along each direction (in Angstroms) as a 3 element vector (sizeZ,sizeY,sizeX).
    Returns:
        (int): 1 if successful and 0 if unsuccessful
    '''
    
    fid = open(filename,'wb')
    
    if len(data.shape) > 3:
        print("Too many dimensions")
        return 0;
    
    if not data.flags['C_CONTIGUOUS']:
        print("Error: Array must be C-style ordering: [numImages,Y,X]. Use numpy.tranpspose and np.ascontiguousarray to change data ordering in memory")
        print('Exiting')
        return 0;
    
    #initialize the header with 256 zeros with size 4 bytes
    header = np.zeros(256,dtype=np.int32)
    fid.write(header)
    fid.seek(0,0) #return to the beginning of the file
    
    #Initialize the int32 part of the header
    header1 = np.zeros(10,dtype=np.int32)
    
    #Write the number of columns, rows and sections (images)
    #header1[0:3] = np.int32(dims) #data size in pixels
    header1[0] = np.int32(data.shape[2]) #num columns, the last index in C-style ordering
    header1[1] = np.int32(data.shape[1]) #num rows
    header1[2] = np.int32(data.shape[0]) #num sections (images)
    
    if data.dtype == np.float32:
        header1[3] = np.int32(2)
    elif data.dtype == np.uint16:
        header1[3] = np.int32(6)
    elif data.dtype == np.int16:
        header1[3] = np.int32(1)
    elif data.dtype == np.int8:
        header1[3] = np.int32(0)
    else:
        print("Data type " + str(data.dtype) + " is unsupported. Only int8, int16, uint16, and float32 are supported")
        return 0;
    
    #Starting point of sub image (not used in IMOD) 
    header1[4:7] = np.zeros(3,dtype=np.int32)
    
    #Grid size in X,Y,Z
    #header1[7:10] = np.int32(dims); #data size in pixels
    header1[7] = np.int32(data.shape[2]) #mx
    header1[8] = np.int32(data.shape[1]) #my
    header1[9] = np.int32(data.shape[0]) #mz
    
    #Write out the first part of the header information
    fid.write(header1)
    
    #Cell dimensions (in Angstroms)
    #pixel spacing = xlen/mx, ylen/my, zlen/mz
    fid.write(np.float32(pixelSize[2]*data.shape[2])) #xlen
    fid.write(np.float32(pixelSize[1]*data.shape[1])) #ylen
    fid.write(np.float32(pixelSize[0]*data.shape[0])) #zlen
    
    #Cell angles (in degrees)
    fid.write(np.float32([90.0,90.0,90.0]))
    
    #Description of array directions with respect to: Columns, Rows, Images
    fid.write(np.int32([1,2,3]))
    
    #Minimum and maximum density
    fid.write(np.float32(np.min(data)))
    fid.write(np.float32(np.max(data)))
    fid.write(np.float32(np.mean(data)))
    
    #Needed to indicate that the data is little endian for NEW-STYLE MRC image2000 HEADER - IMOD 2.6.20 and above
    fid.seek(212,0)
    fid.write(np.int8([68,65,0,0])) #use [17,17,0,0] for big endian
    
    #Write out the data
    fid.seek(1024)
    if forceWrite:    
        fid.write(np.ascontiguousarray(data)) #Change to C ordering array for writing to disk
    else:
        fid.write(data) #msut be C-contiguous

    #Close the file
    fid.close()
    return 1

def writeHeader(filename,shape,dtype,pixelSize):
    '''Write out a MRC type file header according to the specification at http://bio3d.colorado.edu/imod/doc/mrc_format.txt.
    This is useful for initializing an MRC file and then writing to it manually or see appendData() function below.
    
    Parameters:
        filename (str): The name of the EMD file 
        shape (tuple): The shape of the data to write
        pixelSize (tuple): The size of the pixel along each direction (in Angstroms) as a 3 element vector (sizeX,sizeY,sizeZ). sizeZ could be the angular step for a tilt series
    Returns:
        (int): 1 if successful.
    '''
    
    with open(filename,'wb') as fid:
        if len(shape) > 3:
            print("Too many dimensions")
            return 0
        
        #if not data.flags['C_CONTIGUOUS']:
        #    print("Error: Array must be C-style ordering: [numImages,Y,X]. Use np.ascontiguousarray(numpy.transpose(data,[])) to change data ordering in memory\nExiting")
        #    return 0
        
        #Initialize the header with 256 zeros with size 4 bytes
        header = np.zeros(256,dtype=np.int32)
        fid.write(header)
        fid.seek(0,0) #return to the beginning of the file
        
        #Initialize the int32 part of the header with zeros
        header1 = np.zeros(10,dtype=np.int32)
        
        #Write the number of columns, rows and sections (images)
        header1[0] = np.int32(shape[2]) #num columns, the last index in C-style ordering
        header1[1] = np.int32(shape[1]) #num rows
        header1[2] = np.int32(shape[0]) #num sections (images)
        
        if dtype == np.float32:
            header1[3] = np.int32(2)
        elif dtype == np.uint16:
            header1[3] = np.int32(6)
        elif dtype == np.int16:
            header1[3] = np.int32(1)
        elif dtype == np.int8:
            header1[3] = np.int32(0)
        else:
            print("Data type " + str(dtype) + " is unsupported. Only int8, int16, uint16, and float32 are supported")
            return 0;
        
        #Starting point of sub image (not used in IMOD) 
        header1[4:7] = np.zeros(3,dtype=np.int32)
        
        #Grid size in X,Y,Z
        header1[7] = np.int32(shape[2]) #mx
        header1[8] = np.int32(shape[1]) #my
        header1[9] = np.int32(shape[0]) #mz
        
        #Write out the first part of the header information
        fid.write(header1)
        
        #Cell dimensions (in Angstroms)
        #pixel spacing = xlen/mx, ylen/my, zlen/mz
        fid.write(np.float32(pixelSize[2]*shape[2])) #xlen
        fid.write(np.float32(pixelSize[1]*shape[1])) #ylen
        fid.write(np.float32(pixelSize[0]*shape[0])) #zlen
        
        #Cell angles (in degrees)
        fid.write(np.float32([90.0,90.0,90.0]))
        
        #Description of array directions with respect to: Columns, Rows, Images
        fid.write(np.int32([1,2,3]))
        
        #Minimum and maximum density
        #fid.write(np.float32(np.min(data)))
        #fid.write(np.float32(np.max(data)))
        #fid.write(np.float32(np.mean(data)))
        fid.write(np.zeros(3,dtype=np.float32))
        
        #Needed to indicate that the data is little endian for NEW-STYLE MRC image2000 HEADER - IMOD 2.6.20 and above
        fid.seek(212,0)
        fid.write(np.int8([68,65,0,0])) #use [17,17,0,0] for big endian
        
    return 1

def appendData(filename,data):
    '''Append a binary set of data to the end of a MRC file. This should only be used in conjunction with
    writeHeader() above. 
    
    Parameters:
        filename (str):    Name of the MRC file with pre-initiated header and some data already written.
        data (ndarray):   Data to append to the file.
    
    '''
    with open(filename,'ab') as fid:
        #Write out the data
        fid.seek(0,2) #seek to the end of the file
        fid.write(data) #Change to C ordering array for writing to disk
        
def emd2mrc(filename,dsetPath):
    '''Convert EMD data set into MRC data set. The final data type is float32 for convenience.
    
    Parameters:
        filename (str): The name of the EMD file
        dsetPath (str): The HDF5 path to the top group holding the data. ex. '/data/raw/'
    '''
    
    with h5py.File(filename,'r') as f1:
        #Get the pixel sizes and convert to Ang
        dimsPath = dsetPath + '/dim'
        print('Warning: Assuming EMD dim vectors are in nanometer')
        print('dim vector names/units are:')
        for ii in range(1,4):
            print('name, units = {}, {}'.format(f1[dimsPath+str(ii)].attrs['name'],f1[dimsPath+str(ii)].attrs['units']))
        pixelSizeX = (f1[dsetPath + '/dim2'][1] - f1[dsetPath + '/dim2'][0])*10 #change nanometers to Ang
        pixelSizeY = (f1[dsetPath + '/dim3'][1] - f1[dsetPath + '/dim3'][0])*10 #change nanometers to Ang
                
        filenameOut = filename.split('.emd')[0] + '.mrc' #use the first part of the file as the prefix removing the .emd on the end
        
        print('Warning: Converting to float32 before writing to disk')
        mrc.mrcWriter(filenameOut,np.float32(f1[dsetPath+'/data']),(1,pixelSizeY,pixelSizeX)) #the extra slash is not a problem. // is the same as / in a HDF5 data set path
        
        print('Finished writing to: {}'.format(filenameOut))


'''
Tests for the ser io module.
'''

import unittest
import ncempy.io.ser
import numpy as np
import os
import os.path

class test_ser(unittest.TestCase):
    '''
    Test the SER io module.
    '''

    def test_read_ser(self):
        '''
        Test the ser reading functionality.
        '''
        
        # wrong argument type
        with self.assertRaises(TypeError):
            fser = ncempy.io.ser.fileSER(42)

        # non existing file
        with self.assertRaises(IOError):
            fser = ncempy.io.ser.fileSER('')
            
        # wrong file
        with self.assertRaises(Exception):
            fser = ncempy.io.ser.fileSER('ncempy/test/resources/Pt_SAED_D910mm_single/im01.emi', verbose=True)

        # single 2D image file
        fser = ncempy.io.ser.fileSER('ncempy/test/resources/Pt_SAED_D910mm_single/im01_1.ser')
        fser = ncempy.io.ser.fileSER('ncempy/test/resources/Pt_SAED_D910mm_single/im01_1.ser', verbose=True)
        fser = ncempy.io.ser.fileSER('ncempy/test/resources/Pt_SAED_D910mm_single/im01_1.ser', 'ncempy/test/resources/Pt_SAED_D910mm_single/im01.emi', verbose=True)
        
        # time series of 2D images
        fser = ncempy.io.ser.fileSER('ncempy/test/resources/Au_SAED_D910mm_20x_at_800/pos01_1.ser')
        fser = ncempy.io.ser.fileSER('ncempy/test/resources/Au_SAED_D910mm_20x_at_800/pos01_1.ser', verbose=True)
        fser = ncempy.io.ser.fileSER('ncempy/test/resources/Au_SAED_D910mm_20x_at_800/pos01_1.ser','ncempy/test/resources/Au_SAED_D910mm_20x_at_800/pos01.emi', verbose=True)
        
        # mapping of 2D images
        #fser = ncempy.io.ser.fileSER('ncempy/test/resources/CBED_mapping/map01_2.ser', verbose=True)
        
        # mapping of 1D datasets
        #fser = ncempy.io.ser.fileSER('ncempy/test/resources/CBED_mapping/map01_1.ser', verbose=True)
        
        # single 1D dataset
        # time series of 1D datasets
        # 2D mapping of 2D datasets
        # 2D mapping of 1D dataset
        ## not implemented yet
        
        
    def test_read_emi(self):
        '''
        Test the emi reading functionality.
        '''
        
        # ser file for this testing
        fser = ncempy.io.ser.fileSER('ncempy/test/resources/Pt_SAED_D910mm_single/im01_1.ser')
        
        # wrong argument
        with self.assertRaises(TypeError):
            fser.readEMI(42)
        
        # non existing file
        with self.assertRaises(IOError):
            fser.readEMI('')
        
        # auxiliary function
        self.assertIsInstance(fser.parseEntryEMI('42'), int)
        self.assertIsInstance(fser.parseEntryEMI('42.42'), float)
        self.assertIsInstance(fser.parseEntryEMI('forty two'), np.string_)
        
        # read 
        fser.readEMI('ncempy/test/resources/Pt_SAED_D910mm_single/im01.emi')
        
        
    def test_read_dataset(self):
        '''
        Test functions for retrieving datasets.
        '''
    
        # ser file for this testing
        fser = ncempy.io.ser.fileSER('ncempy/test/resources/Pt_SAED_D910mm_single/im01_1.ser')
        
        # wrong index
        with self.assertRaises(IndexError):
            fser.getDataset(-1)
            
        # wrong index type
        with self.assertRaises(TypeError):
            fser.getDataset('foo')
        
        # try dataset + meta
        dataset, meta = fser.getDataset(0, verbose=True)
        
        # try tag
        tag = fser.getTag(0, verbose=True)
        
    
    def test_badtagoffset(self):
        '''
        Bad TagOffsetArray found in some files.
        '''
        
        fser = ncempy.io.ser.fileSER('ncempy/test/resources/ser_badtagoffset/01_Si110_5images_1.ser', 'ncempy/test/resources/ser_badtagoffset/01_Si110_5images.emi', verbose=True)
        
        for i in range(fser.head['ValidNumberElements']):
            tag = fser.getTag(i)
            
            assert(tag['TagTypeID'] == 0)
            assert(tag['Time'] == 0)
            assert(np.isnan(tag['PositionX']))
            assert(np.isnan(tag['PositionY']))
        
        if os.path.isfile('ncempy/test/resources/output/01_Si110_5images.emd'):
            os.remove('ncempy/test/resources/output/01_Si110_5images.emd')
        fser.writeEMD('ncempy/test/resources/output/01_Si110_5images.emd')
        
        #import pdb; pdb.set_trace()
        
        
        
    def test_write_emd(self):
        '''
        Test the emd writing functionality.
        '''
        
        # single 2D image file
        fser = ncempy.io.ser.fileSER('ncempy/test/resources/Pt_SAED_D910mm_single/im01_1.ser')
        if os.path.isfile('ncempy/test/resources/output/Pt_SAED_D910mm_single_woemi.emd'):
            os.remove('ncempy/test/resources/output/Pt_SAED_D910mm_single_woemi.emd')
        fser.writeEMD('ncempy/test/resources/output/Pt_SAED_D910mm_single_woemi.emd')
        fser = ncempy.io.ser.fileSER('ncempy/test/resources/Pt_SAED_D910mm_single/im01_1.ser', 'ncempy/test/resources/Pt_SAED_D910mm_single/im01.emi')
        if os.path.isfile('ncempy/test/resources/output/Pt_SAED_D910mm_single.emd'):
            os.remove('ncempy/test/resources/output/Pt_SAED_D910mm_single.emd')
        fser.writeEMD('ncempy/test/resources/output/Pt_SAED_D910mm_single.emd')
        with self.assertRaises(IOError):
            fser.writeEMD('')
        
        # time series of 2D images
        fser = ncempy.io.ser.fileSER('ncempy/test/resources/Au_SAED_D910mm_20x_at_800/pos01_1.ser', 'ncempy/test/resources/Au_SAED_D910mm_20x_at_800/pos01.emi')
        if os.path.isfile('ncempy/test/resources/output/Au_SAED_D910mm_20x_at_800.emd'):
            os.remove('ncempy/test/resources/output/Au_SAED_D910mm_20x_at_800.emd')
        fser.writeEMD('ncempy/test/resources/output/Au_SAED_D910mm_20x_at_800.emd')
        
        ## exceeds memory
        # mapping of 2D images
        #fser = ncempy.io.ser.fileSER('ncempy/test/resources/CBED_mapping/map01_2.ser')#, 'ncempy/test/resources/CBED_mapping/map01.emi')
        #if os.path.isfile('ncempy/test/resources/output/CBED_mapping_images.emd'):
        #    os.remove('ncempy/test/resources/output/CBED_mapping_images.emd')
        #fser.writeEMD('ncempy/test/resources/output/CBED_mapping_images.emd')
     
        
        # mapping of 1D datasets
        #fser = ncempy.io.ser.fileSER('ncempy/test/resources/CBED_mapping/map01_1.ser', 'ncempy/test/resources/CBED_mapping/map01.emi')
        #if os.path.isfile('ncempy/test/resources/output/CBED_mapping_edx.emd'):
        #    os.remove('ncempy/test/resources/output/CBED_mapping_edx.emd')
        #fser.writeEMD('ncempy/test/resources/output/CBED_mapping_edx.emd')
        
        
        # large time series of 2D images
        fser = ncempy.io.ser.fileSER('ncempy/test/resources/Au_SAED_D910mm_100x_at_RT/step_off_1.ser','ncempy/test/resources/Au_SAED_D910mm_100x_at_RT/step_off.emi', verbose=True)
        ##fser.head['ValidNumberElements'] = 20
        if os.path.isfile('ncempy/test/resources/output/Au_SAED_D910mm_100x_at_RT.emd'):
            os.remove('ncempy/test/resources/output/Au_SAED_D910mm_100x_at_RT.emd')
        fser.writeEMD('ncempy/test/resources/output/Au_SAED_D910mm_100x_at_RT.emd')
        
        # single 1D dataset
        # time series of 1D datasets
        # 2D mapping of 2D datasets
        # 2D mapping of 1D dataset
        ## not implemented yet


# to test with unittest runner
if __name__ == '__main__':
    unittest.main()



'''
Tests for the algo.radial_profile module.
'''

import unittest
import numpy as np
import matplotlib.pyplot as plt
import os

import ncempy.io.emd
import ncempy.algo.local_max
import ncempy.algo.distortion
import ncempy.algo.math
import ncempy.algo.radial_profile


class test_ringdiff(unittest.TestCase):
    '''
    Test the algo.radial_profile module.
    '''
    
    def test_radialprofile(self):
        '''
        Test the radial_profile algorithms to be used on ring diffraction patterns.
        '''
        
        plt.close('all')        
        show=True
        
        # get an image
        femd = ncempy.io.emd.fileEMD('ncempy/test/resources/Pt_SAED_D910mm_single/Pt_SAED_D910mm_single.emd')
        data, dims = femd.get_emdgroup(femd.list_emds[0])
        
        img = data[0,:,:]
        dims = dims[1:3]
        
        # points on a ring
        points = ncempy.algo.local_max.local_max(img, 10, 600)
        points = ncempy.algo.local_max.points_todim(points, dims)
        center_init = ncempy.algo.local_max.points_todim((984,1032), dims)
        points = ncempy.algo.distortion.filter_ring(points, center_init, (6e9, 8e9))
        
        # optimize center
        center = ncempy.algo.distortion.optimize_center(points, center_init, verbose=True)
        
        # fit distortions
        ns = (2,3,4)
        points_plr = ncempy.algo.distortion.points_topolar(points, center)
        dists = ncempy.algo.distortion.optimize_distortion(points_plr, ns)
        
        # check input
        plot = ncempy.algo.distortion.plot_distpolar(points_plr, dims, dists, ns, show=show)
        
        
        ## corrected images
        img_corr = ncempy.algo.radial_profile.correct_distortion( img, dims, center, ns, dists )
        
        
        ## calc_polarcoords
        # wrong input
        with self.assertRaises(TypeError):
            nors, nothes = ncempy.algo.radial_profile.calc_polarcoords( (1,2,3,4,5) )
        with self.assertRaises(TypeError):
            nors, nothes = ncempy.algo.radial_profile.calc_polarcoords( center, dims[0] )
        with self.assertRaises(TypeError):
            nors, nothes = ncempy.algo.radial_profile.calc_polarcoords( center, dims, ns )            
        with self.assertRaises(TypeError):
            nors, nothes = ncempy.algo.radial_profile.calc_polarcoords( center, dims, 42, dists )                    
        with self.assertRaises(TypeError):
            nors, nothes = ncempy.algo.radial_profile.calc_polarcoords( center, dims, ns, dists[0:5] )

        # working
        rs_nodist, thes_nodist = ncempy.algo.radial_profile.calc_polarcoords( center, dims )
        rs, thes = ncempy.algo.radial_profile.calc_polarcoords( center, dims, ns, dists )


        ## calc_radialprofile
        rMax = np.abs(dims[0][0][0]-dims[0][0][1])*np.min(img.shape)/2.0
        dr = np.abs(dims[0][0][0]-dims[0][0][1])/10.
        rsigma = np.abs(dims[0][0][0]-dims[0][0][1])
        # wrong input
        with self.assertRaises(TypeError):
            noR, noI = ncempy.algo.radial_profile.calc_radialprofile( 42, rs, rMax, dr, rsigma )
        with self.assertRaises(TypeError):
            noR, noI = ncempy.algo.radial_profile.calc_radialprofile( img, 42, rMax, dr, rsigma )
        with self.assertRaises(TypeError):
            noR, noI = ncempy.algo.radial_profile.calc_radialprofile( img, rs[0:25,0:25], rMax, dr, rsigma )
        with self.assertRaises(TypeError):
            noR, noI = ncempy.algo.radial_profile.calc_radialprofile( img, rs, 'one', dr, rsigma )
        with self.assertRaises(TypeError):
            noR, noI = ncempy.algo.radial_profile.calc_radialprofile( img, rs, rMax, 'two', rsigma )
        with self.assertRaises(TypeError):
            noR, noI = ncempy.algo.radial_profile.calc_radialprofile( img, rs, rMax, dr, 'three' )
        with self.assertRaises(TypeError):
            noR, noI = ncempy.algo.radial_profile.calc_radialprofile( img, rs, rMax, dr, rsigma, mask=42 )
        with self.assertRaises(TypeError):
            noR, noI = ncempy.algo.radial_profile.calc_radialprofile( img, rs, rMax, dr, rsigma, mask=np.ones((25,25)) )
                                                
        # working
        R, I = ncempy.algo.radial_profile.calc_radialprofile( img, rs, rMax, dr, rsigma )
        
        # everything masked
        emR, emI = ncempy.algo.radial_profile.calc_radialprofile( img, rs, rMax, dr, rsigma, mask=np.zeros(img.shape) )


        ## plot_radialprofile
        # wrong input
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.radial_profile.plot_radialprofile( 42, I, dims )
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.radial_profile.plot_radialprofile( R, 42, dims )
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.radial_profile.plot_radialprofile( R, I, 'notdims' )
        
        # working
        plot = ncempy.algo.radial_profile.plot_radialprofile( R, I, dims, show=show) 
        plot = ncempy.algo.radial_profile.plot_radialprofile( emR, emI, dims, show=show)
        
        
        ## compare without distortion correction
        R_nodist, I_nodist = ncempy.algo.radial_profile.calc_radialprofile( img, rs_nodist, rMax, dr, rsigma )
        plot = ncempy.algo.radial_profile.plot_radialprofile( R_nodist, I_nodist, dims, show=show)
        
        
        # cut radial profile
        sel = (R>=1.5e9)*(R<=9.5e9)
        I = I[sel]
        R = R[sel]
        

        ## fit_radialprofile
        funcs = [ 'const', 'powlaw', 'voigt' ]
        init_guess = (  10,
                        1.0e12, -1.0,
                        5e10, 7.3e9, 1.1e7, 2.5e7 )
        
        # wrong input
        with self.assertRaises(TypeError):
            nores = ncempy.algo.radial_profile.fit_radialprofile( 42, I, funcs, init_guess )
        with self.assertRaises(TypeError):
            nores = ncempy.algo.radial_profile.fit_radialprofile( R, 42, funcs, init_guess )
        with self.assertRaises(TypeError):
            nores = ncempy.algo.radial_profile.fit_radialprofile( R, I[0:25], funcs, init_guess )
        with self.assertRaises(TypeError):
            nores = ncempy.algo.radial_profile.fit_radialprofile( R, I, funcs, init_guess[0:3] )
        with self.assertRaises(TypeError):
            nores = ncempy.algo.radial_profile.fit_radialprofile( R, I, 42, init_guess )
        
        # working
        res = ncempy.algo.radial_profile.fit_radialprofile( R, I, funcs, init_guess, maxfev=10000 ) 
                 
        
        ## plot_fit
        # wrong input
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.radial_profile.plot_fit( 42, I, dims, funcs, res )
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.radial_profile.plot_fit( R, 42, dims, funcs, res )
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.radial_profile.plot_fit( R, I, 'notdims', funcs, res )
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.radial_profile.plot_fit( R, I, dims, funcs, res[0:3] )
        with self.assertRaises(TypeError):
            noplot = ncempy.algo.radial_profile.plot_fit( R, I, dims, 42, res )

        # before fit
        plot = ncempy.algo.radial_profile.plot_fit( R, I, dims, funcs, init_guess, show=show )
        # after fit
        plot = ncempy.algo.radial_profile.plot_fit( R, I, dims, funcs, res, show=show )
        
        
        # try two voigts
        funcs = [ 'voigt', 'voigt' ]
        init_guess = ( 6e10, 7.3e9, 2e7, 2e7, 
                       8e10, 2.4e9, 1e8, 1e8 )
        res = ncempy.algo.radial_profile.fit_radialprofile( R, I, funcs, init_guess, maxfev=10000 )
        plot = ncempy.algo.radial_profile.plot_fit( R, I, dims, funcs, res, show=show )


        # subtract a power law background fitted to specific points
        # get some fixpoints
        fit_xs = [1.3e9, 1.5e9, 9.05e9]
        fit_xswidth = 0.05e9
        fit_R = np.array([])
        fit_I = np.array([])
        for xpoint in fit_xs:
            ix = np.where( np.abs(R-xpoint) < fit_xswidth )
            fit_R = np.append(fit_R, R[ix])
            fit_I = np.append(fit_I, I[ix])
            
        funcs = [ 'const', 'powlaw']
        init_guess = (1, 1.0e12, -1.0)
        res = ncempy.algo.radial_profile.fit_radialprofile( fit_R, fit_I, funcs, init_guess, maxfev=1000 )
        plot = ncempy.algo.radial_profile.plot_fit( R, I, dims, funcs, res, show=show )
        
        I = I - ncempy.algo.math.sum_functions(R, funcs, res)
        
        plot = ncempy.algo.radial_profile.plot_radialprofile( R, I, dims, show=show )

        #import pdb;pdb.set_trace()
        
        # wait for plots
        if show:
            plt.show()            
    

# to test with unittest runner
if __name__ == '__main__':
    unittest.main()



'''
Tests for the basic functionalities of the dm io module.
'''

import datetime
import inspect
import ncempy.io.dm
import os
import unittest
from matplotlib.image import imread


def get_measure(start=None, operation=None):
    '''Auxiliary function to measure time spans.
    
    Args:
        start: if set to a datetime object, function returns a timespan
          object measurring the difference between start and now().
        operation: if set to a string, (and start is a datetime object),
          the function prints a message indicating the process PID, the
          string and the time_sona value.
    
    Returns: a datetime object (now()) or a timespan object.    
    '''
    if start is None:
        return  datetime.datetime.now()
    else:
        end = datetime.datetime.now()
        delta = end-start
        if operation:
            print("PID({})-Operation ({}): {}".format(os.getpid(), operation,
                                                  delta))
        return delta

class test_dm3(unittest.TestCase):
    '''
    Test the DM3 io module
    '''
    
    ''' Auxiliary attributes and method to facilitate the location of the test
    images'''
    _curren_file = os.path.dirname(os.path.abspath(
                    inspect.getfile(inspect.currentframe())))
    _images_folder="{}/resources".format(_curren_file)

    def _get_image_route(self, file_name):
        return os.path.join(self._images_folder, file_name)
          
    
    def _read_dm3_data(self, file_route, on_memory=False):
        '''Creates a DMobject and reads its data metadata
        
        Args:
             file_route: full route where the file is placed.
             on_memory: if True, the dm file will be opened in on memory mode.
            
        
        '''
        
        f = ncempy.io.dm.fileDM(file_route, on_memory=on_memory)
        if on_memory:
            self.assertTrue(f._on_memory)
        f.parseHeader()
        ds = f.getDataset(0)
        img3D = ds['data']
        dimensions=len(img3D.shape);
        
        img = img3D
        if dimensions == 3:
            img = img[:,:,int(img.shape[2]/2),]
        elif dimensions == 4:
            img = img[:,:,
                  int(img.shape[2]/2),
                  int(img.shape[3]/2)]
        elif dimensions>4:
            raise ValueError("Images with more than four dimensions not"
                            " supported.")
        metadata = dict(dimensions=dimensions,
                              header=f.allTags,
                              metadata={x:ds[x]
                                        for x in ds.keys()
                                        if x!="data"})
        del f
        return metadata, img
    
    def test_read_dm3(self):
        
        metadata, img = self._read_dm3_data(self._get_image_route(
                            "dmTest_3D_float32_nonSquare_diffPixelSize.dm3"))

        self.assertEqual(metadata["dimensions"], 3)
        self.assertEqual(metadata["metadata"]["pixelOrigin"], [0.0, 0.0, 0.0])
        self.assertEqual(metadata["header"]['.ImageSourceList.1.ImageRef'], 1)
        
    def test_read_dm3_one_dimension(self):
        
        metadata, img = self._read_dm3_data(self._get_image_route(
                            "06_lowLoss.dm3"))

        self.assertEqual(metadata["dimensions"], 2)
        self.assertEqual(metadata["metadata"]["pixelOrigin"], [0.0, 200.0])
        self.assertEqual(metadata["header"]['.ImageSourceList.1.ImageRef'], 1)

        
    
    def test_read_dm3_on_memory(self):
        
        metadata, img = self._read_dm3_data(self._get_image_route(
                            "dmTest_3D_float32_nonSquare_diffPixelSize.dm3"),
                            on_memory=True)
        
        
        self.assertEqual(metadata["dimensions"], 3)
        self.assertEqual(metadata["metadata"]["pixelOrigin"], [0.0, 0.0, 0.0])
        self.assertEqual(metadata["header"]['.ImageSourceList.1.ImageRef'], 1)
        
    def test_dm4_memory_vs_file_performance(self):
        """ Even with a local HD, memory read should be x10 faster."""
        m0=get_measure()
        for i in range(10):
            metadata, img = self._read_dm3_data(
                             self._get_image_route(
                                 "Si-SiGe-test-01-31x12x448x480.dm4"))
            delta0=get_measure(m0)
            
            m1=get_measure()
            metadata, img = self._read_dm3_data(
                                self._get_image_route(
                                    "Si-SiGe-test-01-31x12x448x480_copy.dm4"),
                                on_memory=True)
            delta1=get_measure(m1)
            
            self.assertGreater(delta0, delta1)
        
    def test_extract_on_memory(self):
        from matplotlib.image import imsave
        from matplotlib import cm
        f = ncempy.io.dm.fileDM(
            self._get_image_route("Si-SiGe-test-01-31x12x448x480.dm4"),
            on_memory=False)
        
        f.parseHeader()
        ds = f.getDataset(0)
        img3D_no_on_mem = ds['data']
        
        del f
        
        f = ncempy.io.dm.fileDM(
            self._get_image_route("Si-SiGe-test-01-31x12x448x480.dm4"),
            on_memory=True)
        
        f.parseHeader()
        ds = f.getDataset(0)
        img3D_on_mem = ds['data']
      
        del f
        self.assertTrue((img3D_no_on_mem==img3D_on_mem).all())
        
    def text_compare_png(self):
        
        for file_name in ["dmTest_3D_float32_nonSquare_diffPixelSize.dm3",
                          "dmTest_uint16_nonSquare_diffPixelSize.dm3",
                          "dmTest_int32_nonSquare_diffPixelSize.dm4"]:
            metadata, img = self._read_dm3_data(self._get_image_route(
                                "{}".format(file_name)),
                                on_memory=True)
            png = imread(self._get_image_route(
                            "{}.png".format(file_name)))
            self.assertEaual(img, png)
        
        


'''
Module to find local maxima in an image.
'''

import numpy as np
import scipy.ndimage.filters

import matplotlib.pyplot as plt


def local_max(img, r, thresh):
    '''Find local maxima from comparing dilated and eroded images.
    
    Calculates images with maximum and minimum within given radius. If the difference is larger than the threshold, the original pixel position with max value is detected as local maximum.
    
    Parameters:
        img (np.ndarray):    Input image.
        r (int):    Radius for locality.
        thresh (int/float):    Intensity difference threshold.
    
    Returns:
        (np.ndarray):    Array of points.
        
    '''
    
    try:
        r = int(r)
        thresh = int(thresh)
        assert(isinstance(img, np.ndarray))
    except:
        raise TypeError('Bad input!')
    
    
    # prepare circular kernel
    y,x = np.ogrid[-r:r+1, -r:r+1]
    kernel = x**2 + y**2 <= r**2
    
    # calculate max and min images
    img_dil = scipy.ndimage.filters.maximum_filter(img, footprint=kernel)
    img_ero = scipy.ndimage.filters.minimum_filter(img, footprint=kernel)
    
    # get selection of local maxima
    sel = (img==img_dil)*(img-img_ero > thresh)
    
    if sel.any():  
        # retrieve and return points
        points = np.argwhere(sel)
        #points = np.roll(points, 1, axis=1)
        
        return points
    else:
        # otherwise return None to avoid having an empty list
        return None
        
    
def plot_points(img, points, vminmax=(0,1), dims=None, invert=False, show=False):
    '''Plot the detected points on the input image for checking.
    
    Parameters:
        img (np.ndarray):    Image.
        points (np.ndarray):    Array containing the points.
        vminmax (tuple):    Tuple of two values for relative lower and upper cut off to display image.
        dims (tuple):    Tuple of dims to plot in dimensions.
        invert (bool):    Set to invert the image.
        show (bool):    Set to directly show the plot interactively.
    
    Returns:
        (np.ndarray):    Image of the plot.
        
    '''
    
    try:
        assert(isinstance(img, np.ndarray))
        
        assert(isinstance(points, np.ndarray))
        assert(points.shape[1] == 2)
        assert(len(points.shape) == 2)
    except:
        raise TypeError('Something wrong with the input!')
    
    fig = plt.figure()
    ax = fig.add_subplot(111)
    
    if invert:
        cmap="Greys"
    else:
        cmap="gray"
    
    if dims:
        ax.imshow(img, cmap=cmap, vmin=np.min(img)+vminmax[0]*(np.max(img)-np.min(img)), vmax=np.min(img)+vminmax[1]*(np.max(img)-np.min(img)), extent=(np.min(dims[0][0]), np.max(dims[0][0]), np.max(dims[1][0]), np.min(dims[1][0])) )
        ax.set_xlabel('{} {}'.format(dims[0][1], dims[0][2]))
        ax.set_ylabel('{} {}'.format(dims[1][1], dims[1][2]))
        ax.set_xlim((np.min(dims[0][0]), np.max(dims[0][0])))
        ax.set_ylim((np.max(dims[1][0]), np.min(dims[1][0])))
    else:
        ax.imshow(img, cmap=cmap, vmin=np.min(img)+vminmax[0]*(np.max(img)-np.min(img)), vmax=np.min(img)+vminmax[1]*(np.max(img)-np.min(img)) )
        ax.set_xlim((0,img.shape[1]-1))
        ax.set_ylim((img.shape[0]-1,0))
    
    ax.scatter(points[:,1], points[:,0], color='r', marker='o', facecolors='none')
    
    if show:
        try:
            plt.show(block=False)
        except:
            pass
    
    fig.canvas.draw()
    
    plot = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    plot = plot.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    
    return plot
    
    
def points_todim(points, dims):
    '''Convert points from px coordinates to real dim.
    
    Points are expected to be array indices for the first two dimensions in dims.
    
    Parameters:
        points (np.ndarray):    Points to convert.
        dims (tuple):    Tuple of dimensions.
        
    Returns:
        (np.ndarray):   Converted points.
        
    '''
    
    try:
        # try to convert input to np.ndarray with 2 columns (necessary if only one entry provided)
        points = np.reshape(np.array(points), (-1,2))
        # check if enough dims availabel
        assert(len(dims)>=2)
        assert(len(dims[0])==3)
    except:
        raise TypeError('Something wrong with the input!')
    
    # do the conversion by looking up thing in dimension vectors
    points_d = np.array( [ dims[0][0][points[:,0]], dims[1][0][points[:,1]] ] ).transpose()
    
    return points_d



'''
Tests for the eva.ring_diff module.
'''

import unittest
import numpy as np
import matplotlib.pyplot as plt
import os

import ncempy.io.emd
import ncempy.algo.local_max
import ncempy.algo.math
import ncempy.eval.ring_diff


class test_ringdiff(unittest.TestCase):
    '''
    Test the evaluation of ring diffraction patterns.
    '''
    
    def test_eva(self):
        
        plt.close('all')        
        show=True
        
        
        ### test settings writing and reading
        if os.path.isfile('ncempy/test/resources/output/test_settings.emd'):
            os.remove('ncempy/test/resources/output/test_settings.emd')
            
        femd = ncempy.io.emd.fileEMD('ncempy/test/resources/output/test_settings.emd')
        
        ## minimum settings necessary, with many nones
        settings_minimum = { 'lmax_r': 10,
                 'lmax_thresh': 600,
                 'lmax_cinit': (984, 1032),
                 'lmax_range': (6e9, 8e9),
                 'plt_imgminmax': None,
                 'ns': (2,3,4),
                 'rad_rmax': None,
                 'rad_dr': None,
                 'rad_sigma': None,
                 'mask': None,
                 'fit_rrange': (1.5e9, 9.5e9),
                 'back_xs': (1.3e9, 1.5e9, 9.05e9),
                 'back_xswidth': 0.05e9,
                 'back_init': (1, 1.0e12, -1.0),
                 'fit_funcs': ('voigt',),
                 'fit_init': ( 5e10, 7.3e9, 1.1e7, 2.5e7 ),
                 'fit_maxfev': None
               }
            
        # put       
        par = femd.file_hdl.create_group('put_minimum')
        
        # wrong input
        with self.assertRaises(TypeError):
            ncempy.eval.ring_diff.put_settings( 'not_par', settings_minimum )
        with self.assertRaises(TypeError):
            ncempy.eval.ring_diff.put_settings( par, 42 )
              
        ncempy.eval.ring_diff.put_settings( par, settings_minimum )
        
        # get
        
        # wrong input
        with self.assertRaises(TypeError):
            no_settings = ncempy.eval.ring_diff.get_settings('not a group')        
        
        in_settings_minimum = ncempy.eval.ring_diff.get_settings( par['settings_ringdiffraction'] )
        
        # compare the two settings
        for key in settings_minimum:
            self.assertTrue(key in in_settings_minimum)
            
            comp = (settings_minimum[key]==in_settings_minimum[key])
            if isinstance(comp, bool):
                self.assertTrue(comp)
            else:
                self.assertTrue(comp.all())
        
        ## full settings with all parameters set
        settings_full = { 'lmax_r': 10,
                 'lmax_thresh': 600,
                 'lmax_cinit': (984, 1032),
                 'lmax_range': (6e9, 8e9),
                 'plt_imgminmax': (0.0, 0.2),
                 'ns': (2,3,4),
                 'rad_rmax': 42.,
                 'rad_dr': 0.1,
                 'rad_sigma': 0.01,
                 'mask': np.ones((25,25)),
                 'fit_rrange': (1.5e9, 9.5e9),
                 'back_xs': (1.3e9, 1.5e9, 9.05e9),
                 'back_xswidth': 0.05e9,
                 'back_init': (1, 1.0e12, -1.0),
                 'fit_funcs': ('voigt',),
                 'fit_init': ( 5e10, 7.3e9, 1.1e7, 2.5e7 ),
                 'fit_maxfev': 1000
               }
               
        # put       
        par = femd.file_hdl.create_group('put_full')
        ncempy.eval.ring_diff.put_settings( par, settings_full )
        
        # get
        in_settings_full = ncempy.eval.ring_diff.get_settings( par['settings_ringdiffraction'] )
        
        # compare the two settings
        for key in settings_full:
            self.assertTrue(key in in_settings_full)
            
            comp = (settings_full[key]==in_settings_full[key])
            if isinstance(comp, bool):
                self.assertTrue(comp)
            else:
                self.assertTrue(comp.all())
        
        
        ### single evaluation
        femd = ncempy.io.emd.fileEMD('ncempy/test/resources/Pt_SAED_D910mm_single/Pt_SAED_D910mm_single.emd')
        emdgrp = femd.list_emds[0]
        
        if os.path.isfile('ncempy/test/resources/output/evaluation_Pt_SAED_D910mm_single.emd'):
            os.remove('ncempy/test/resources/output/evaluation_Pt_SAED_D910mm_single.emd')
        femd_out = ncempy.io.emd.fileEMD('ncempy/test/resources/output/evaluation_Pt_SAED_D910mm_single.emd')
        
        # write evaluation details
        grp_eva = femd_out.file_hdl.create_group('evaluation')
        
        # wrong input
        with self.assertRaises(TypeError):
            nohdl = ncempy.eval.ring_diff.put_sglgroup('nogroup', 'not_working', emdgrp)
        with self.assertRaises(TypeError):
            nohdl = ncempy.eval.ring_diff.put_sglgroup(grp_eva, 'not_working', 'nogroup')
        
        hdl = ncempy.eval.ring_diff.put_sglgroup(grp_eva, 'Pt_SAED_D910mm_single', emdgrp)

        # put the settings
        settings = { 'lmax_r': 10,
                 'lmax_thresh': 600,
                 'lmax_cinit': (984, 1032),
                 'lmax_range': (6e9, 8e9),
                 'plt_imgminmax': (0.,0.2),
                 'ns': (2,3,4),
                 'rad_rmax': None,
                 'rad_dr': None,
                 'rad_sigma': None,
                 'mask': None,
                 'fit_rrange': (1.5e9, 9.5e9),
                 'back_xs': (1.3e9, 1.5e9, 9.05e9),
                 'back_xswidth': 0.05e9,
                 'back_init': (1, 1.0e12, -1.0),
                 'fit_funcs': ('voigt',),
                 'fit_init': ( 5e10, 7.3e9, 1.1e7, 2.5e7 ),
                 'fit_maxfev': None
               }
        ncempy.eval.ring_diff.put_settings(femd_out.file_hdl, settings)

        # run the evaluation
        # wrong input
        with self.assertRaises(TypeError):
            ncempy.eval.ring_diff.run_sglgroup(42, femd_out)
        with self.assertRaises(TypeError):
            ncempy.eval.ring_diff.run_sglgroup(hdl, 'notanemdfile')
                    
        ncempy.eval.ring_diff.run_sglgroup(hdl, femd_out, verbose=True, showplots=False)
        
        
        ### evaluation of a series
        femd = ncempy.io.emd.fileEMD('ncempy/test/resources/Au_SAED_D910mm_20x_at_800/Au_SAED_D910mm_20x_at_800.emd')
        emdgrp = femd.list_emds[0]
        
        if os.path.isfile('ncempy/test/resources/output/evaluation_Au_SAED_D910mm_20x_at_800.emd'):
            os.remove('ncempy/test/resources/output/evaluation_Au_SAED_D910mm_20x_at_800.emd')
        femd_out = ncempy.io.emd.fileEMD('ncempy/test/resources/output/evaluation_Au_SAED_D910mm_20x_at_800.emd')
        
        # write evaluation details
        grp_eva = femd_out.file_hdl.create_group('evaluation')
        
        hdl = ncempy.eval.ring_diff.put_sglgroup(grp_eva, 'Au_SAED_D910mm_20x_at_800', emdgrp)
        
        # put the settings
        settings = { 'lmax_r': 16,
                     'lmax_thresh': 250,
                     'lmax_cinit': (1012, 1020),
                     'lmax_range': (6.5e9, 7.3e9),
                     'plt_imgminmax': (0.,0.4),
                     'ns': (2,3,4),
                     'rad_rmax': None,
                     'rad_dr': None,
                     'rad_sigma': None,
                     'mask': None,
                     'fit_rrange': (1.5e9, 9.5e9),
                     'back_xs': (1.5e9, 5.7e9, 9.05e9),
                     'back_xswidth': 0.05e9,
                     'back_init': (1, 1.5e9, -0.8),
                     'fit_funcs': ('voigt','voigt','voigt','voigt','voigt'),
                     'fit_init': ( 3.4e10, 4.25e9, 1.1e7, 2.5e7,
                                   9.1e9, 4.90e9, 1.1e7, 2.5e7,
                                   1.6e10, 6.95e9, 1.1e7, 2.5e7,
                                   1.1e10, 8.14e9, 1.1e7, 2.5e7,
                                   1.6e9, 8.50e9, 1.1e7, 2.5e7 ),
                     'fit_maxfev': None
                   }
        ncempy.eval.ring_diff.put_settings(femd_out.file_hdl, settings)

        # run the evaluation
        ncempy.eval.ring_diff.run_sglgroup(hdl, femd_out, verbose=True, showplots=False)
        
        
        ### run multiple evaluations
        ### single evaluation
        femd = ncempy.io.emd.fileEMD('ncempy/test/resources/Pt_SAED_D910mm_single/Pt_SAED_D910mm_single.emd')
        emdgrp = femd.list_emds[0]
        
        if os.path.isfile('ncempy/test/resources/output/evaluation_Pt_SAED_D910mm_multiple.emd'):
            os.remove('ncempy/test/resources/output/evaluation_Pt_SAED_D910mm_multiple.emd')
        femd_out = ncempy.io.emd.fileEMD('ncempy/test/resources/output/evaluation_Pt_SAED_D910mm_multiple.emd')
        
        # write evaluation details
        grp_eva = femd_out.file_hdl.create_group('evaluation')
        hdl = ncempy.eval.ring_diff.put_sglgroup(grp_eva, 'Pt_SAED_D910mm_first', emdgrp)
        hdl = ncempy.eval.ring_diff.put_sglgroup(grp_eva, 'Pt_SAED_D910mm_second', emdgrp)
        hdl = ncempy.eval.ring_diff.put_sglgroup(grp_eva, 'Pt_SAED_D910mm_third', emdgrp)

        # put the settings
        settings = { 'lmax_r': 10,
                 'lmax_thresh': 600,
                 'lmax_cinit': (984, 1032),
                 'lmax_range': (6e9, 8e9),
                 'plt_imgminmax': (0.,0.2),
                 'ns': (2,3,4),
                 'rad_rmax': None,
                 'rad_dr': None,
                 'rad_sigma': None,
                 'mask': None,
                 'fit_rrange': (1.5e9, 9.5e9),
                 'back_xs': (1.3e9, 1.5e9, 9.05e9),
                 'back_xswidth': 0.05e9,
                 'back_init': (1, 1.0e12, -1.0),
                 'fit_funcs': ('voigt',),
                 'fit_init': ( 5e10, 7.3e9, 1.1e7, 2.5e7 ),
                 'fit_maxfev': None
               }
        ncempy.eval.ring_diff.put_settings(grp_eva, settings)
        
        ncempy.eval.ring_diff.run_all(grp_eva, femd_out, verbose=True, showplots=False)
        

        # wait for plots
        if show:
            plt.show()     
        

# to test with unittest runner
if __name__ == '__main__':
    unittest.main()



'''
This module provides an interface to the SER file format written by TIA.

Following the information provided by Dr Chris Boothroyd (http://www.er-c.org/cbb/info/TIAformat/).
'''
import numpy as np
import h5py
import os
import re
import xml.etree.ElementTree as ET
import datetime

class NotSERError(Exception):
    '''Exception if a file is not in SER file format.
    
    '''
    pass


class fileSER:
    '''
    Class to represent SER files (read only).
    
    Parameters:
        filename (str):    Name of the SER file.
        
        emifile (str):    Name of an optional _emi file to read metadata.
        
        verbose (bool):    True to get extensive output while reading the file.
    
    Examples:
        Simple example for reading data from a single image into memory:
        
        >>> import matplotlib.pyplot as plt
        >>> from ncempy.io import dm
        >>> with dm.fileDM('filename.dm4') as dmFile1:
            dataSet = dmFile1.getDataset(0)
        >>> plt.imshow(dataSet['data'])

        SER files are structures such that each image in a series is a different dataset.
        Thus, time series data should be read as the following:
        
        >>> with ser.fileSER('filename_1.ser') as ser1:
            image0,metadata0 = ser1.getDataset(0)
            image1,metadata1 = ser1.getDataset(1)
        >>> plt.imshow(image)
        >>> print('Pixel size for dimension 0 = {} meters'.format(metadata['Calibration'][0]['CalibrationDelta']))
    '''

    _dictByteOrder = {0x4949 : 'little endian'}
    '''(dict):    Information on byte order.'''
    
    _dictSeriesVersion = {0x0210 : '< TIA 4.7.3', 0x0220 : '>= TIA 4.7.3'}
    '''(dict):    Information on fileformat version.'''
    
    _dictDataTypeID = {0x4120:'1D datasets', 0x4122:'2D images'}
    '''(dict):    Information on data type.'''
    
    _dictTagTypeID = {0x4152:'time only',0x4142:'time and 2D position'}
    '''(dict):    Information on tag type.'''
    
    _dictDataType = {1:'<u1', 2:'<u2', 3:'<u4', 4:'<i1', 5:'<i2', 6:'<i4', 7:'<f4', 8:'<f8', 9:'<c8', 10:'<c16'}
    '''(dict):    Information on data format.'''


    def __init__(self, filename, emifile=None, verbose=False):
        '''Init opening the file and reading in the header.
        
        '''
        # necessary declarations, if something fails
        self._file_hdl = None
        self._emi = None

        # check for string
        if not isinstance(filename, str):
            raise TypeError('Filename is supposed to be a string')

        # try opening the file
        try:
            self._file_hdl = open(filename, 'rb')
        except IOError:
            print('Error reading file: "{}"'.format(filename))
            raise
        except :
            raise

        # read header
        self.head = self.readHeader(verbose)
        
        # read _emi, if provided
        if emifile:
            self._emi = self.read_emi(emifile)

    def __del__(self):
        '''Closing the file stream on del.
        
        '''
        
        # close the file
        if(not self._file_hdl.closed):
            self._file_hdl.close()

    def __enter__(self):
        '''Implement python's with staement
        
        '''
        return self
        
    def __exit__(self,type,value,traceback):
        '''Implement python's with statment
        and close the file via __del__()
        '''
        self.__del__()
        return None
        
    def readHeader(self, verbose=False):
        '''Read and return the SER files header.
        
        Parameters:
            verbose (bool):    True to get extensive output while reading the file.

        Returns:
            (dict):    The header of the SER file as dict.
            
        '''
        
        # prepare empty dict to be populated while reading
        head = {}

        # go back to beginning of file
        self._file_hdl.seek(0,0)

        # read 3 int16
        data = np.fromfile(self._file_hdl, dtype='<i2', count=3)

        # ByteOrder (only little Endian expected)
        if not data[0] in self._dictByteOrder:
            raise RuntimeError('Only little Endian implemented for SER files')
        head['ByteOrder'] = data[0]
        if verbose:
            print('ByteOrder:\t"{:#06x}",\t{}'.format(data[0], self._dictByteOrder[data[0]]))

        # SeriesID, check whether TIA Series Data File   
        if not data[1] == 0x0197:
            raise NotSERError('This is not a TIA Series Data File (SER)')
        head['SeriesID'] = data[1]
        if verbose:
            print('SeriesID:\t"{:#06x},\tTIA Series Data File'.format(data[1]))

        # SeriesVersion
        if not data[2] in self._dictSeriesVersion:
            raise RuntimeError('Unknown TIA version: "{:#06x}"'.format(data[2]))
        head['SeriesVersion'] = data[2]
        if verbose:
            print('SeriesVersion:\t"{:#06x}",\t{}'.format(data[2], self._dictSeriesVersion[data[2]]))
        # version dependend fileformat for below
        if head['SeriesVersion']==0x0210:
            offset_dtype = '<i4'
        else: #head['SeriesVersion']==0x220:
            offset_dtype = '<i8'
        
        # read 4 int32
        data = np.fromfile(self._file_hdl, dtype='<i4', count=4)

        # DataTypeID
        if not data[0] in self._dictDataTypeID:
            raise RuntimeError('Unknown DataTypeID: "{:#06x}"'.format(data[0]))
        head['DataTypeID'] = data[0]
        if verbose:
            print('DataTypeID:\t"{:#06x}",\t{}'.format(data[0], self._dictDataTypeID[data[0]]))

        # TagTypeID
        if not data[1] in self._dictTagTypeID:
            raise RuntimeError('Unknown TagTypeID: "{:#06x}"'.format(data[1]))
        head['TagTypeID'] = data[1]
        if verbose:
            print('TagTypeID:\t"{:#06x}",\t{}'.format(data[1], self._dictTagTypeID[data[1]]))

        # TotalNumberElements
        if not data[2] >= 0:
            raise RuntimeError('Negative total number of elements: {}'.format(data[2]))
        head['TotalNumberElements'] = data[2]
        if verbose:
            print('TotalNumberElements:\t{}'.format(data[2]))

        # ValidNumberElements
        if not data[3] >= 0:
            raise RuntimeError('Negative valid number of elements: {}'.format(data[3]))
        head['ValidNumberElements'] = data[3]
        if verbose:
            print('ValidNumberElements:\t{}'.format(data[3]))
        
        # OffsetArrayOffset, sensitive to SeriesVersion
        data = np.fromfile(self._file_hdl, dtype=offset_dtype, count=1)
        head['OffsetArrayOffset'] = data[0]
        if verbose:
            print('OffsetArrayOffset:\t{}'.format(data[0]))

        # NumberDimensions
        data = np.fromfile(self._file_hdl, dtype='<i4', count=1)
        if not data[0] >= 0:
            raise RuntimeError('Negative number of dimensions')
        head['NumberDimensions']=data[0]
        if verbose:
            print('NumberDimensions:\t{}'.format(data[0]))


        # Dimensions array
        dimensions = []
        for i in range(head['NumberDimensions']):
            if verbose:
                print('reading Dimension {}'.format(i))
            this_dim = {}
        
            # DimensionSize
            data = np.fromfile(self._file_hdl, dtype='<i4', count=1)
            this_dim['DimensionSize'] = data[0]
            if verbose:
                print('DimensionSize:\t{}'.format(data[0]))
            
            data = np.fromfile(self._file_hdl, dtype='<f8', count=2)
            
            # CalibrationOffset
            this_dim['CalibrationOffset'] = data[0]
            if verbose:
                print('CalibrationOffset:\t{}'.format(data[0]))
            
            # CalibrationDelta
            this_dim['CalibrationDelta'] = data[1]
            if verbose:
                print('CalibrationDelta:\t{}'.format(data[1]))
            
            data = np.fromfile(self._file_hdl, dtype='<i4', count=2)
            
            # CalibrationElement
            this_dim['CalibrationElement'] = data[0]
            if verbose:
                print('CalibrationElement:\t{}'.format(data[0]))
            
            # DescriptionLength
            n = data[1]
            
            # Description
            data = np.fromfile(self._file_hdl, dtype='<i1', count=n)
            data = ''.join(map(chr, data))
            this_dim['Description'] = data
            if verbose:
                print('Description:\t{}'.format(data))
            
            # UnitsLength
            data = np.fromfile(self._file_hdl, dtype='<i4', count=1)
            n = data[0]
            
            # Units
            data = np.fromfile(self._file_hdl, dtype='<i1', count=n)
            data = ''.join(map(chr, data))
            this_dim['Units'] = data
            if verbose:
                print('Units:\t{}'.format(data))


            dimensions.append(this_dim)
        
        # save dimensions array as tuple of dicts in head dict
        head['Dimensions'] = tuple(dimensions)
        
        
        # Offset array
        self._file_hdl.seek(head['OffsetArrayOffset'],0)
        
        # DataOffsetArray
        data = np.fromfile(self._file_hdl, dtype=offset_dtype, count=head['ValidNumberElements'])
        head['DataOffsetArray'] = data.tolist()
        if verbose:
            print('reading in DataOffsetArray')
        
        # TagOffsetArray
        data = np.fromfile(self._file_hdl, dtype=offset_dtype, count=head['ValidNumberElements'])
        head['TagOffsetArray'] = data.tolist()
        if verbose:
            print('reading in TagOffsetArray')     

        return head


    def _checkIndex(self, i):
        '''Check index i for sanity, otherwise raise Exception.
        
        Parameters:
            i (int):    Index.
            
        '''
        
        # check type
        if not isinstance(i, int):
            raise TypeError('index supposed to be integer')

        # check whether in range
        if i < 0 or i>= self.head['ValidNumberElements']:
            raise IndexError('Index out of range, trying to access element {} of {} valid elements'.format(i+1, self.head['ValidNumberElements']))
            
        return
        

    def getDataset(self, index, verbose=False):
        '''Retrieve dataset from data file.

        Parameters:
            index (int):    Index of dataset.
            verbose (bool):    True to get extensive output while reading the file.
        
        Returns:
            (tuple):    Tuple containing:
            
                np.ndarray:    Dataset as array.
                
                dict:    Metadata as dict.
                
        '''

        # check index, will raise Exceptions if not
        try:
            self._checkIndex(index)
        except:
            raise

        if verbose:
            print('Getting dataset {} of {}.'.format(index, self.head['ValidNumberElements']))
            
        # go to dataset in file
        self._file_hdl.seek(self.head['DataOffsetArray'][index],0)
        
        # read meta
        meta = {}
        
        # number of calibrations depends on DataTypeID
        if self.head['DataTypeID'] == 0x4120:
            n = 1
        elif self.head['DataTypeID'] == 0x4122:
            n = 2
        else:
            raise RuntimeError('Unknown DataTypeID')
       
        # read in the calibrations    
        cals = []
        for i in range(n):
            if verbose:
                print('Reading calibration {}'.format(i))
                
            this_cal = {}
        
            data = np.fromfile(self._file_hdl, dtype='<f8', count=2)
            
            # CalibrationOffset
            this_cal['CalibrationOffset'] = data[0]
            if verbose:
                print('CalibrationOffset:\t{}'.format(data[0]))
            
            # CalibrationDelta
            this_cal['CalibrationDelta'] = data[1]
            if verbose:
                print('CalibrationDelta:\t{}'.format(data[1]))
            
            data = np.fromfile(self._file_hdl, dtype='<i4', count=1)
            
            # CalibrationElement
            this_cal['CalibrationElement'] = data[0]
            if verbose:
                print('CalibrationElement:\t{}'.format(data[0]))
            
            cals.append(this_cal)
            
        meta['Calibration'] = tuple(cals)
        
        data = np.fromfile(self._file_hdl, dtype='<i2', count=1)
        
        # DataType
        meta['DataType'] = data[0]
        
        if not data[0] in self._dictDataType:
            raise RuntimeError('Unknown DataType: "{}"'.format(data[0]))
        if verbose:
            print('DataType:\t{},\t{}'.format(data[0],self._dictDataType[data[0]]))
                
        if self.head['DataTypeID'] == 0x4120:
            # 1D data element
        
            data = np.fromfile(self._file_hdl, dtype='<i4', count=1)
            # ArrayLength
            data = data.tolist()
            meta['ArrayShape'] = data
            if verbose:
                print('ArrayShape:\t{}'.format(data))
                
            dataset = np.fromfile(self._file_hdl, dtype=self._dictDataType[meta['DataType']], count=meta['ArrayShape'][0])
        
        elif self.head['DataTypeID'] == 0x4122:
            # 2D data element
            
            data = np.fromfile(self._file_hdl, dtype='<i4', count=2)
            # ArrayShape
            data = data.tolist()
            meta['ArrayShape'] = data
            if verbose:
                print('ArrayShape:\t{}'.format(data))

            # dataset
            dataset = np.fromfile(self._file_hdl, dtype=self._dictDataType[meta['DataType']], count=meta['ArrayShape'][0]*meta['ArrayShape'][1])
            dataset = dataset.reshape(meta['ArrayShape'][::-1]) #needs to be reversed for little endian data
        
            dataset = np.flipud(dataset)

        return dataset, meta


    def _getTag(self, index, verbose=False):
        '''Retrieve tag from data file.

        Parameters:
            index (int):    Index of tag.
            verbose (bool):    True to get extensive output while reading the file.

        Returns:
            (dict):    Tag as dict.
            
        '''

        # check index, will raise Exceptions if not
        try:
            self._checkIndex(index)
        except:
            raise

        if verbose:
            print('Getting tag {} of {}.'.format(index, self.head['ValidNumberElements']))
        
        tag = {}
            
        try:
            # bad tagoffsets occured pointing to the end of the file
        
            # go to dataset in file
            self._file_hdl.seek(self.head['TagOffsetArray'][index],0)
            
            data = np.fromfile(self._file_hdl, dtype='<i4', count=2)
        
            # TagTypeID
            tag['TagTypeID'] = data[0]
            
            # only proceed if TagTypeID is the same like in the file header (bad TagOffsetArray issue)
            if tag['TagTypeID'] == head['TagTypeID']:
                if verbose:
                    print('TagTypeID:\t"{:#06x}",\t{}'.format(data[0], self._dictTagTypeID[data[0]]))
                    
                # Time    
                tag['Time'] = data[1]
                if verbose:
                    print('Time:\t{}'.format(data[1]))
                
                # check for position
                if tag['TagTypeID'] == 0x4142:
                    data = np.fromfile(self._file_hdl, dtype='<f8', count=2)
                    
                    # PositionX
                    tag['PositionX'] = data[0]
                    if verbose:
                        print('PositionX:\t{}'.format(data[0]))
                    
                    # PositionY
                    tag['PositionY'] = data[1]
                    if verbose:
                        print('PositionY:\t{}'.format(data[1]))
            else:
                # otherwise raise to get to default tag
                raise
                    
        except:
            tag['TagTypeID'] = 0
            tag['Time'] = 0
            tag['PositionX'] = np.nan
            tag['PositionY'] = np.nan
        
     
        return tag
        
    
    def _createDim(self, size, offset, delta, element):
        '''Create dimension labels from SER information.
        
        Parameters:
            size (int):    Number of elements.
            offset (float):    Value at indicated element.
            delta (float):    Difference between elements.
            element (int):    Indicates the element of value offset.
        
        Returns:
            (np.ndarray):    Dimension vector as array.
        
        '''
        
        #dim = np.zeros(size)
        #dim[:] = np.nan
        
        # if element is out off range, map it back into defined
        if element >= size:
            element = size-1
            offset = offset - (element-(size-1))*delta
        
        dim = np.array(range(size)).astype('f8')
        dim = dim*delta
        dim += (offset - dim[element])
        
        # some weird shifting, positionx is +0.5, positiony is -0.5
        # doing this during saving
        # dim += 0.5*delta
        
        return dim
    
    
    def _parseEntry_emi(self, value):
        '''Auxiliary function to parse string entry to int, float or np.string_().
        
        Parameters:
            value (str):    String containing an int, float or string.
        
        Returns:
            (int/float/str):    Value as int, float or string.
            
        '''
        
        # try to parse as int
        try:
            p = int(value)
        except:
            # if not int, then try float
            try:
                p = float(value)
            except:
                # if neither int nor float, stay with string
                p = np.string_(str(value))
        
        return p
        
    
    def read_emi(self, filename):
        '''Read the meta data from an _emi file.
        
        Parameters:
            filename (str):    Name of the _emi file.
        
        Returns:
            (dict):    Dict of metadata.
        '''
        
        # check for string
        if not isinstance(filename, str):
            raise TypeError('Filename is supposed to be a string')

        # try opening the file
        try:
            # open file for reading bytes, as binary and text are intermixed
            f_emi = open(filename, 'rb')
        except IOError:
            print('Error reading file: "{}"'.format(filename))
            raise
        except :
            raise
        
        # dict to store _emi stuff
        _emi = {}
        
        # need anything readable from <ObjectInfo> to </ObjectInfo>
        collect = False
        data = b''
        for line in f_emi:
            if b'<ObjectInfo>' in line:
                collect = True
            if collect:
                data += line.strip()
            if b'</ObjectInfo>' in line:
                collect = False

        # close the file
        f_emi.close()
            
        # strip of binary stuff still around
        data = data.decode('ascii', errors='ignore')
        matchObj = re.search('<ObjectInfo>(.+?)</ObjectInfo', data)
        try:
            data = matchObj.group(1)
        except:
            raise RuntimeError('Could not find _emi metadata in specified file.')

        # parse metadata as xml
        root = ET.fromstring('<_emi>'+data+'</_emi>')
        
        # single items
        _emi['Uuid'] = root.findtext('Uuid')
        _emi['AcquireDate'] = root.findtext('AcquireDate')
        _emi['Manufacturer'] = root.findtext('Manufacturer')
        _emi['DetectorPixelHeigth'] = root.findtext('DetectorPixelHeight')
        _emi['DetectorPixelWidth'] = root.findtext('DetectorPixelWidth')

        # Microscope Conditions
        grp = root.find('ExperimentalConditions/MicroscopeConditions')
        
        for elem in grp:
            _emi[elem.tag] = self._parseEntry_emi(elem.text)
            #print('{}:\t{}'.format(elem.tag, elem.text)) 

        # Experimental Description
        grp = root.find('ExperimentalDescription/Root')
        
        for elem in grp:
            _emi['{} [{}]'.format(elem.findtext('Label'), elem.findtext('Unit'))] = self._parseEntry_emi(elem.findtext('Value'))
            #print('{} [{}]: \t{}'.format(elem.findtext('Label'), elem.findtext('Unit'), elem.findtext('Value')))

        # AcquireInfo
        grp = root.find('AcquireInfo')
        
        for elem in grp:
            _emi[elem.tag] = self._parseEntry_emi(elem.text)
            
        # DetectorRange
        grp = root.find('DetectorRange')
        
        for elem in grp:
            _emi['DetectorRange_'+elem.tag] = self._parseEntry_emi(elem.text)

        return _emi
        
        
    def writeEMD(self, filename):
        '''
        Write SER data to an EMD file.
        
        Parameters:
            filename (str):    Name of the EMD file.
        '''
        
        from ncempy.io import emd
        
        # create the EMD file and set version attributes
        try:
            f = emd.fileEMD(filename)
        except:
            raise IOError('Cannot write to file "{}"!'.format(filename))
        
        # create EMD group    
        grp = f._file_hdl['data'].create_group(os.path.basename(self._file_hdl.name))
        grp.attrs['emd_group_type'] = 1
            
        
        # use first dataset to layout memory
        data, first_meta = self.getDataset(0)
        first_tag = self._getTag(0)
        
        if self.head['DataTypeID'] == 0x4122:
            # 2D datasets
            self.head['ExperimentType'] = 'image' #text indicator of the experiment type
            if first_tag['TagTypeID'] == 0x4142:
                # 2D mapping
                dset = grp.create_dataset('data', (self.head['Dimensions'][1]['DimensionSize'], self.head['Dimensions'][0]['DimensionSize'], first_meta['ArrayShape'][1], first_meta['ArrayShape'][0]), dtype=self._dictDataType[first_meta['DataType']] )
                
                # collect time
                time = np.zeros( (self.head['Dimensions'][0]['DimensionSize'], self.head['Dimensions'][1]['DimensionSize']), dtype='i4')

                # create mapping dims for checking
                map_xdim = self._createDim( self.head['Dimensions'][0]['DimensionSize'], self.head['Dimensions'][0]['CalibrationOffset'], self.head['Dimensions'][0]['CalibrationDelta'], self.head['Dimensions'][0]['CalibrationElement'] )
                map_ydim = self._createDim( self.head['Dimensions'][1]['DimensionSize'], self.head['Dimensions'][1]['CalibrationOffset'], self.head['Dimensions'][1]['CalibrationDelta'], self.head['Dimensions'][1]['CalibrationElement'] )
                # weird direction dependend half pixel shifting
                map_xdim += 0.5*self.head['Dimensions'][0]['CalibrationDelta']
                map_ydim -= 0.5*self.head['Dimensions'][1]['CalibrationDelta']

                for y in range(self.head['Dimensions'][0]['DimensionSize']):
                    for x in range(self.head['Dimensions'][1]['DimensionSize']):

                        index = int(x+y*self.head['Dimensions'][0]['DimensionSize'])
                        print('converting dataset {} of {}, items ({}, {})'.format(index+1, self.head['ValidNumberElements'], x, y))
               
                        # retrieve dataset and put into buffer     
                        data, meta = self.getDataset( index )
                        dset[y, x, :,:] = data[:,:]
                        
                        # get tag data per image
                        tag = self._getTag(index)
                        time[y,x] = tag['Time']

                        assert( np.abs(tag['PositionX'] - map_xdim[x]) < np.abs(tag['PositionX']*1e-8) )
                        assert( np.abs(tag['PositionY'] - map_ydim[y]) < np.abs(tag['PositionY']*1e-8) )
                
                        del data, meta, tag
                
                # create dimension datasets
                dims = []
                dims_time = []
                
                # Position Y
                assert self.head['Dimensions'][1]['Description'] == 'Position'
                dims.append( (map_ydim, self.head['Dimensions'][1]['Description'], '[{}]'.format(self.head['Dimensions'][1]['Units'])) )
                dims_time.append( (map_ydim, self.head['Dimensions'][1]['Description'], '[{}]'.format(self.head['Dimensions'][1]['Units'])) ) 
                
                # Position X
                assert self.head['Dimensions'][0]['Description'] == 'Position'
                dims.append( (map_xdim, self.head['Dimensions'][0]['Description'], '[{}]'.format(self.head['Dimensions'][0]['Units'])) )
                dims_time.append( (map_xdim, self.head['Dimensions'][0]['Description'], '[{}]'.format(self.head['Dimensions'][0]['Units'])) )
                
                dim = self._createDim(first_meta['ArrayShape'][1], first_meta['Calibration'][1]['CalibrationOffset'], first_meta['Calibration'][1]['CalibrationDelta'], first_meta['Calibration'][1]['CalibrationElement'])
                dims.append( (dim, 'y', '[m]') )
                
                dim = self._createDim(first_meta['ArrayShape'][0], first_meta['Calibration'][0]['CalibrationOffset'], first_meta['Calibration'][0]['CalibrationDelta'], first_meta['Calibration'][0]['CalibrationElement'])
                dims.append( (dim, 'x', '[m]') )
                
                # write dimensions
                for i in range(len(dims)):
                    f.write_dim('dim{:d}'.format(i+1), dims[i], grp)

                # write out time as additional dataset
                grp = f.put_emdgroup('timestamp', time, dims_time, parent=grp)
                
                
            else:
            
                # 1 entry series to single image
                if self.head['ValidNumberElements'] == 1:
                
                    # get image
                    data, meta = self.getDataset(0)
                    tag = self._getTag(0)
                    
                    # create dimensions
                    dims = []
                    
                    dim = self._createDim(first_meta['ArrayShape'][1], first_meta['Calibration'][1]['CalibrationOffset'], first_meta['Calibration'][1]['CalibrationDelta'], first_meta['Calibration'][1]['CalibrationElement'])
                    dims.append( (dim, 'y', '[m]') )
                                                                      
                    dim = self._createDim(first_meta['ArrayShape'][0], first_meta['Calibration'][0]['CalibrationOffset'], first_meta['Calibration'][0]['CalibrationDelta'], first_meta['Calibration'][0]['CalibrationElement'])
                    dims.append( (dim, 'x', '[m]') )
                    
                    dset = grp.create_dataset( 'data', (first_meta['ArrayShape'][1], first_meta['ArrayShape'][0]), dtype=self._dictDataType[first_meta['DataType']])
                    
                    dset[:,:] = data[:,:]
                    
                    for i in range(len(dims)):
                        f.write_dim('dim{:d}'.format(i+1), dims[i], grp)
                        
                    dset.attrs['timestamp'] = tag['Time']
                    
                else:
            
                    # simple series
                    dset = grp.create_dataset( 'data', (self.head['ValidNumberElements'], first_meta['ArrayShape'][1], first_meta['ArrayShape'][0]), dtype=self._dictDataType[first_meta['DataType']] )
        
                    # collect time
                    time = np.zeros(self.head['ValidNumberElements'], dtype='i4')
        
                    for i in range(self.head['ValidNumberElements']):
                        print('converting dataset {} of {}'.format(i+1, self.head['ValidNumberElements']))
            
                        # retrieve dataset and put into buffer
                        data, meta = self.getDataset(i)
                        dset[i,:,:] = data[:,:]
            
                        # get tag data per image
                        tag = self._getTag(i)
                        time[i] = tag['Time']
                        
                    # create dimension datasets
                    dims = []
                        
                    # first SER dimension is number
                    assert self.head['Dimensions'][0]['Description'] == 'Number'
                        
                    dim = self._createDim(self.head['Dimensions'][0]['DimensionSize'], self.head['Dimensions'][0]['CalibrationOffset'], self.head['Dimensions'][0]['CalibrationDelta'], self.head['Dimensions'][0]['CalibrationElement'])
                    dims.append( (dim[0:self.head['ValidNumberElements']], self.head['Dimensions'][0]['Description'], '[{}]'.format(self.head['Dimensions'][0]['Units'])) )

                    dim = self._createDim(first_meta['ArrayShape'][1], first_meta['Calibration'][1]['CalibrationOffset'], first_meta['Calibration'][1]['CalibrationDelta'], first_meta['Calibration'][1]['CalibrationElement'])
                    dims.append( (dim, 'y', '[m]') )
                                                                      
                    dim = self._createDim(first_meta['ArrayShape'][0], first_meta['Calibration'][0]['CalibrationOffset'], first_meta['Calibration'][0]['CalibrationDelta'], first_meta['Calibration'][0]['CalibrationElement'])
                    dims.append( (dim, 'x', '[m]') )

                    # write dimensions
                    for i in range(len(dims)):
                        f.write_dim('dim{:d}'.format(i+1), dims[i], grp)
                    
                    # write out time as additional dim vector
                    f.write_dim('dim1_time', (time, 'timestamp', '[s]'), grp)
        
        
        elif self.head['DataTypeID'] == 0x4120:
            # 1D datasets; spectra
            self.head['ExperimentType'] = 'spectrum' #text indicator of the experiment type
            
            if first_tag['TagTypeID'] == 0x4142:
                    # 2D mapping
                    dset = grp.create_dataset( 'data', (self.head['Dimensions'][1]['DimensionSize'], self.head['Dimensions'][0]['DimensionSize'], first_meta['ArrayShape'][0]), dtype=self._dictDataType[first_meta['DataType']] )
                    
                    time = np.zeros( (self.head['Dimensions'][0]['DimensionSize'], self.head['Dimensions'][1]['DimensionSize']), dtype='i4')

                    # create mapping dims for checking
                    map_xdim = self._createDim( self.head['Dimensions'][0]['DimensionSize'], self.head['Dimensions'][0]['CalibrationOffset'], self.head['Dimensions'][0]['CalibrationDelta'], self.head['Dimensions'][0]['CalibrationElement'] )
                    map_ydim = self._createDim( self.head['Dimensions'][1]['DimensionSize'], self.head['Dimensions'][1]['CalibrationOffset'], self.head['Dimensions'][1]['CalibrationDelta'], self.head['Dimensions'][1]['CalibrationElement'] )
                    # weird direction dependend half pixel shifting
                    map_xdim += 0.5*self.head['Dimensions'][0]['CalibrationDelta']
                    map_ydim -= 0.5*self.head['Dimensions'][1]['CalibrationDelta']

                    for y in range(self.head['Dimensions'][0]['DimensionSize']):
                        for x in range(self.head['Dimensions'][1]['DimensionSize']):

                            index = int(x+y*self.head['Dimensions'][0]['DimensionSize'])
                            print('converting dataset {} of {}, items ({}, {})'.format(index+1, self.head['ValidNumberElements'], x, y))
                   
                            # retrieve dataset and put into buffer     
                            data, meta = self.getDataset( index )
                            dset[y, x, :] = np.copy(data[:])
                            
                            # get tag data per image
                            tag = self._getTag(index)
                            time[y,x] = tag['Time']
                                
                            assert( np.abs(tag['PositionX'] - map_xdim[x]) < np.abs(tag['PositionX']*1e-8) )
                            assert( np.abs(tag['PositionY'] - map_ydim[y]) < np.abs(tag['PositionY']*1e-8) )
                    
                            del data, meta, tag
                    
                    # create dimension datasets
                    dims = []
                    dims_time = []
                    
                    # Position Y
                    assert self.head['Dimensions'][1]['Description'] == 'Position'
                    dims.append( (map_ydim, self.head['Dimensions'][1]['Description'], '[{}]'.format(self.head['Dimensions'][1]['Units'])) )
                    dims_time.append( (map_ydim, self.head['Dimensions'][1]['Description'], '[{}]'.format(self.head['Dimensions'][1]['Units'])) )
                    
                    # Position X
                    assert self.head['Dimensions'][0]['Description'] == 'Position'
                    dims.append( (map_xdim, self.head['Dimensions'][0]['Description'], '[{}]'.format(self.head['Dimensions'][0]['Units'])) )
                    dims_time.append( (map_xdim, self.head['Dimensions'][0]['Description'], '[{}]'.format(self.head['Dimensions'][0]['Units'])) )
                    
                    dim = self._createDim(first_meta['ArrayShape'][0], first_meta['Calibration'][0]['CalibrationOffset'], first_meta['Calibration'][0]['CalibrationDelta'], first_meta['Calibration'][0]['CalibrationElement'])
                    dims.append( (dim, 'E', '[m_eV]') )
                        
                    # write dimensions
                    for i in range(len(dims)):
                        f.write_dim('dim{:d}'.format(i+1), dims[i], grp)
 
                    # write out time as additional dataset
                    grp = f.put_emdgroup('timestamp', time, dims_time, parent=grp)
                    
            else:
                    # simple series
                    dset = grp.create_dataset( 'data', (self.head['ValidNumberElements'], first_meta['ArrayShape'][0]), dtype=self._dictDataType[first_meta['DataType']])
                    
                    # collect time
                    time = np.zeros(self.head['ValidNumberElements'], dtype='i4')
                    
                    for i in range(self.head['ValidNumberElements']):
                        print('converting dataset {} of {}'.format(i+1, self.head['ValidNumberElements']))
            
                        # retrieve dataset and put into buffer
                        data, meta = self.getDataset(i)
                        dset[i,:] = data[:]
            
                        # get tag data per image
                        tag = self._getTag(i)
                        time[i] = tag['Time']

                    # create dimension datasets
                    dims = []

                    # first SER dimension is number
                    assert self.head['Dimensions'][0]['Description'] == 'Number'
                    dim = self._createDim(self.head['Dimensions'][0]['DimensionSize'], self.head['Dimensions'][0]['CalibrationOffset'], self.head['Dimensions'][0]['CalibrationDelta'], self.head['Dimensions'][0]['CalibrationElement'])
                    dims.append( (dim[0:self.head['ValidNumberElements']], self.head['Dimensions'][0]['Description'], '[{}]'.format(self.head['Dimensions'][0]['Units'])) )
                        
                    dim = self._createDim(first_meta['ArrayShape'][0], first_meta['Calibration'][0]['CalibrationOffset'], first_meta['Calibration'][0]['CalibrationDelta'], first_meta['Calibration'][0]['CalibrationElement'])
                    dims.append( (dim, 'E', '[m_eV]') )
                        
                    # write dimensions
                    for i in range(len(dims)):
                        f.write_dim('dim{:d}'.format(i+1), dims[i], grp)
 
                    # write out time as additional dim vector
                    f.write_dim('dim1_time', (time, 'timestamp', '[s]'), grp)    
                    
        else:
            raise RuntimeError('Unknown DataTypeID')    
            

        # attempt to free memory asap
        #del dset_buf, dims
        
        
        # put meta information from _emi to Microscope group, if available
        if self._emi:
            for key in self._emi:
                if not self._emi[key] is None:
                    f.microscope.attrs[key] = self._emi[key]
                
        # write comment into Comment group
        f.put_comment('Converted SER file "{}" to EMD using the openNCEM tools.'.format(self._file_hdl.name))
        
def serReader(filename):
    '''Simple function to parse the file and read all datasets. This is a one function implementation to load all data in a ser file.
    
    Parameters:
        filename (str) : The filename containing the data
        
    Returns:
        (dict) : A dictionary containing the data and meta data. The data is in the key 'data'
        
    Examples:
        Load a single image data set and show the image:
            
            >>> from ncempy.io import ser
            >>> im0 = ser.serReader('filename_1.ser')
            >>> plt.imshow(im0['data']) #show the single image from the data file
    '''
    with fileSER(filename) as f1: #open the file and init the class
            
        if f1.head['ValidNumberElements'] > 0:
            data, metaData = f1.getDataset(0) #get the first data set to setup the arrays
            
            metaData['filename'] = filename #save the file name in the output dictionary
            
            npType = f1._dictDataType[metaData['DataType']]
            
            if f1.head['DataTypeID'] == 0x4120:
                #spectra as 1D single spectra, 2D line scan or 3D spectrum image
                numSpectra = f1.head['ValidNumberElements']
                spectraSize = data.shape[0]
                
                #Read in all spectra
                temp = np.zeros((numSpectra,spectraSize),dtype=npType) #C-style ordering
                for ii in range(0,numSpectra):
                    dataValues,meta1 = f1.getDataset(ii)
                    temp[ii,:] = dataValues
                
                if f1.head['NumberDimensions'] > 1:
                    #Spectrum map
                    scanI = f1.head['Dimensions'][0]['DimensionSize']
                    scanJ = f1.head['Dimensions'][1]['DimensionSize']
                    temp = temp.reshape((scanJ,scanI,spectraSize)) #operations on spectra are fastest
                else:
                    temp = np.squeeze(temp)
                    
                #Setup the energy loss axis for convenience
                eDelta = metaData['Calibration'][0]['CalibrationDelta']
                eOffset = metaData['Calibration'][0]['CalibrationOffset']
                eLoss = np.linspace(0,(spectraSize-1)*eDelta,spectraSize) + eOffset
                
                dataOut = {'data':temp,'eLoss':eLoss,'eOffset':eOffset,'eDelta':eDelta,'scanCalibration':f1.head['Dimensions']}
            elif f1.head['DataTypeID'] == 0x4122:
                #images as 2D or 3D image series
                temp = np.empty([f1.head['ValidNumberElements'],data.shape[0],data.shape[1]],dtype=npType)
                for ii in range(0,f1.head['ValidNumberElements']):
                    data0,metadata0 = f1.getDataset(ii)
                    temp[ii,:,:] = data #get the next dataset
                    
                    if ii ==1:
                        metadata = metadata0
                temp = np.squeeze(temp) #remove singular dimensions
                
                #Get the metadata from the first image
                data,metadata = f1.getDataset(0)
                
                dataOut = {}
                dataOut['data'] = temp #output the full data set
                
                #Setup some simple meta data
                dataOut['pixelSize'] = []
                dataOut['pixelUnit'] = []
                dataOut['pixelOrigin'] = []
                
                for cal in metadata['Calibration']:
                    dataOut['pixelSize'].append(cal['CalibrationDelta'])
                    dataOut['pixelOrigin'].append(cal['CalibrationOffset'])
                    dataOut['pixelUnit'].append('m')
                dataOut['filename'] = filename #save the file name
                
        else:
            dataOut = {}
            print('No data set found') 
        #del f1 #delete the class and close the file
    return dataOut #return the dataset and metadata as a dictionary


'''
Tests for the algo.local_max module.
'''

import unittest
import numpy as np
import matplotlib.pyplot as plt
import os

import ncempy.io.emd
import ncempy.algo.local_max


class test_localmax(unittest.TestCase):
    '''
    Test the localmax module on diffraction patterns.
    '''
    
    def test_local_max(self):
        '''
        Test the local_maxima algorithm to be used for ring diffraction patterns.
        '''
        
        plt.close('all')        
        show=True
        
        # get an image
        femd = ncempy.io.emd.fileEMD('ncempy/test/resources/Pt_SAED_D910mm_single/Pt_SAED_D910mm_single.emd')
        data, dims = femd.get_emdgroup(femd.list_emds[0])
        
        img = data[0,:,:]
        dims = dims[1:3]
        
        
        ## local_max
        # not working
        with self.assertRaises(TypeError):
            points = ncempy.algo.local_max.local_max(42, 10, 600)
        with self.assertRaises(TypeError):
            points = ncempy.algo.local_max.local_max(img, 'ten', 600)
        with self.assertRaises(TypeError):
            points = ncempy.algo.local_max.local_max(img, 10, 'sixhundred')
            
        # no points detected
        self.assertIsNone (ncempy.algo.local_max.local_max(img, 10, 100000))
        
        # working
        points = ncempy.algo.local_max.local_max(img, 10, 600)
        self.assertIsNotNone(points)
        
        
        ## plot_points
        # not working
        with self.assertRaises(TypeError):
            plot = ncempy.algo.local_max.plot_points(42, points)
        with self.assertRaises(TypeError):
            plot = ncempy.algo.local_max.plot_points(img, np.array([[0,1,2,3,4],[5,6,7,8,9]]))
            
        # px coords, non inverted
        plot = ncempy.algo.local_max.plot_points(img, points, vminmax=(0.0,0.2), show=show)
        
        
        ## points_todim
        # not working
        with self.assertRaises(TypeError):
            points = ncempy.algo.local_max.points_todim([0,1,2,3,5], dims)
        with self.assertRaises(TypeError):
            points = ncempy.algo.local_max.points_todim(points, dims[0])
        
        # convert points to real dim
        point = ncempy.algo.local_max.points_todim((0,0), dims)
        points = ncempy.algo.local_max.points_todim(points, dims)
        
        # real coords, inverted
        plot = ncempy.algo.local_max.plot_points(img, points, vminmax=(0.0,0.2), dims=dims, invert=True, show=show)
        
        
        # wait for plots
        if show:
            plt.show()
    
    
# to test with unittest runner
if __name__ == '__main__':
    unittest.main()



'''
Module to correlate two images, functionally written.

'''

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
#import cv2

def multicorr(G1, G2, method = 'cross', upsampleFactor = 1):
    '''Align a template to an image, possibly with multiple alignment peaks.
    
    Inputs:
        G1 - Fourier transform of image we are aligning to. Image reference
        G2 - Fourier transform of image that is being aligned. Image being registered
    Optional Inputs:
        'phase' or 'cross' or 'hybrid' - string that specifies correlation method used. (default = 'cross')
        upsampleFactor - scalar integer specifying 1/subpixel_precision of fit. (default = 1)

    Returns:
        xyShift - the shift between G1 and G2 in pixels

    Defaults -
        method = 'phase';
        peakSearchMode = false;  By default, we return only global best shift value.
        upsampleFactor = 1;
    '''
    method, upsampleFactor = parse_input(G1, G2, method, upsampleFactor)
    imageCorr = initial_correlation_image(G1, G2, method, upsampleFactor)
    xyShift = upsampled_correlation(imageCorr, upsampleFactor)
    print(method)
    print('------------')
    return xyShift

def parse_input(G1, G2, method = 'cross', upsampleFactor = 1):
    '''This function parses the inputs to make sure they're correct. Will raise errors if G1 or G2 are not ndarrays.

    Inputs:
        G1 - Fourier transform of image we are aligning to. Image reference
        G2 - Fourier transform of image that is being aligned. Image being registered
        method - correlation method ('phase', 'cross', 'hybrid'). Default 'cross'
        upsampleFactor - scalar integer specifying 1/subpixel_precision of fit. Default = 1.

    Returns:
        method - string ('phase', 'cross', 'hybrid')
        upsampleFactor - scalar integer
    '''
    # Check to make sure both G1 and G2 are arrays
    if type(G1) is not np.ndarray:
        raise TypeError('G1 must be an ndarray')
    elif type(G2) is not np.ndarray:
        raise TypeError('G2 must be an ndarray')

    # Check to make sure method and upsample factor are the correct values
    if method not in ['phase', 'cross', 'hybrid']:
        print('Unknown method used, setting to cross')
        method = 'cross'

    if type(upsampleFactor) is not int and type(upsampleFactor) is not float:
        print('Upsample factor is not an integer or float, setting to 1')
        upsampleFactor = 1
    elif type(upsampleFactor) is not int:
        print('Upsample factor is not an integer, rounding down')
        upsampleFactor = int(upsampleFactor)
        if upsampleFactor < 1:
            print('Upsample factor is < 1, setting to 1')
            upsampleFactor = 1

    # Verify images are the same size.
    if G1.shape != G2.shape:
        raise TypeError('G1 and G2 are not the same size, G1 is {0} and G2 is {1}'.format(str(G1.shape), str(G2.shape)))

    return method, upsampleFactor

def initial_correlation_image(G1, G2, method = 'cross', upsampleFactor = 1):
    '''Generate correlation image at initial resolution using the method specified.

    Inputs:
        G1 - Fourier transform of image we are aligning to. Image reference
        G2 - Fourier transform of image that is being aligned. Image being registered
        method - correlation method ('phase', 'cross', 'hybrid'). Default 'cross'
        upsampleFactor - scalar integer specifying 1/subpixel_precision of fit. Default = 1.

    Returns:
        imageCorr - an ndarray correlation image that has not yet been inverse Fourier transformed.
    '''
    G12 = np.multiply(G2, np.conj(G1)) # is this the correct order that we want?
    if method == 'phase':
        imageCorr = np.exp(1j * np.angle(G12))
    elif method == 'cross':
        imageCorr = G12
    elif method == 'hybrid':
        imageCorr = np.multiply(np.sqrt(np.absolute(G12)), np.exp(1j * np.angle(G12)))
    else:
        raise TypeError('{} method is not allowed'.format(str(method)))

    return imageCorr

def upsampled_correlation(imageCorr, upsampleFactor):
    '''Upsamples the correlation image by a set integer factor upsampleFactor. If upsampleFactor == 2, then it is naively Fourier upsampled. If the upsampleFactoris higher than 2, then it uses dftUpsample, which is a more efficient way to Fourier upsample the image.

    Inputs:
        imageCorr - Fourier transformed correlation image returned by initial_correlation_image. Is an ndarray.
        upsampleFactor - scalar integer of how much upsampling should be performed.

    Returns:
        xyShift - two element list with shift in x and y of G2 with respect to G1.
    '''

    imageCorrIFT = np.real(np.fft.ifft2(imageCorr))
    xyShift = list(np.unravel_index(imageCorrIFT.argmax(), imageCorrIFT.shape, 'C'))
    # print(['xyShift pre mod '] + xyShift)
    if upsampleFactor == 1:
        imageSize = imageCorrIFT.shape
        xyShift[0] = ((xyShift[0] + imageSize[0]/2) % imageSize[0]) - imageSize[0]/2
        xyShift[1] = ((xyShift[1] + imageSize[1]/2) % imageSize[1]) - imageSize[1]/2
        # print(['xyShift post mod '] + xyShift)
        #G2shift = np.fft.fft2(np.roll(np.roll(np.fft.ifft2(G2), int(xyShift[0]), 0), int(xyShift[1]), 1))
    else:
        imageCorrLarge = upsampleFFT(imageCorr, 2)
        imageSizeLarge = imageCorrLarge.shape
        xySubShift2 = list(np.unravel_index(imageCorrLarge.argmax(), imageSizeLarge, 'C'))
        print(['xySubShift2 '] + xySubShift2)
        xySubShift2[0] = ((xySubShift2[0] + imageSizeLarge[0]/2) % imageSizeLarge[0]) - imageSizeLarge[0]/2
        xySubShift2[1] = ((xySubShift2[1] + imageSizeLarge[1]/2) % imageSizeLarge[1]) - imageSizeLarge[1]/2
        xyShift = [i/2 for i in xySubShift2] #signs have to flip, or mod wrong?
        # print(xySubShift2)
        print(['xyShiftln127'] + xyShift)

        if upsampleFactor > 2:
            # here is where we use DFT registration to make things much faster
            # we cut out and upsample a peak 1.5 by 1.5 px from our original correlation image.

            xyShift[0] = np.round(xyShift[0] * upsampleFactor) / upsampleFactor
            xyShift[1] = np.round(xyShift[1] * upsampleFactor) / upsampleFactor

            globalShift = np.fix(np.ceil(upsampleFactor * 1.5)/2)# this line might have an off by one error based. The associated matlab comment is "this will be used to center the output array at dftshift + 1"
            print('globalShift', globalShift, 'upsampleFactor', upsampleFactor, 'xyShift', xyShift)

            imageCorrUpsample = np.conj(dftUpsample(np.conj(imageCorr), upsampleFactor, globalShift - np.multiply(xyShift, upsampleFactor))) / (np.fix(imageSizeLarge[0]) * np.fix(imageSizeLarge[1]) * upsampleFactor ** 2)

            xySubShift = np.unravel_index(imageCorrUpsample.argmax(), imageCorrUpsample.shape, 'C')
            # xySubShift = np.add(list(xySubShift), [1, 1])
            print('xySubShift', xySubShift)

            # add a subpixel shift via parabolic fitting
            try:
                icc = np.real(imageCorrUpsample[xySubShift[0] - 1 : xySubShift[0] + 2, xySubShift[1] - 1 : xySubShift[1] + 2])
                dx = (icc[2,1] - icc[0,1]) / (4 * icc[1,1] - 2 * icc[2,1] - 2 * icc[0,1])
                dy = (icc[1,2] - icc[1,0]) / (4 * icc[1,1] - 2 * icc[1,2] - 2 * icc[1,0])
            except:
                dx, dy = 0, 0 # this is the case when the peak is near the edge and one of the above values does not exist
            print('dxdy', dx, dy)
            print('xyShift', xyShift)
            xySubShift = xySubShift - globalShift;
            print('xysubShift2', xySubShift)
            xyShift = xyShift + (xySubShift + np.array([dx, dy])) / upsampleFactor
            print('xyShift2', xyShift)

    return xyShift

def upsampleFFT(imageInit, upsampleFactor):
    '''This does a Fourier upsample of the imageInit. imageInit is the Fourier transform of the correlation image. upsampleFactor is self-descriptive. The function returns the real space correlation image that has been Fourier upsampled by 2. It is written generally such that upsampleFactor can be greater than 2, but that should never happend/it has not been tested.

    The way it works is that it embeds imageInit in a larger array of zeros, then does the inverse Fourier transform to return the Fourier upsampled image in real space.

    Inputs:
        imageInit - ndarray of the image to be Fourier upsampled. This should be in the Fourier domain.
        upsampleFactor - integer scalar, almost always 2.

    Returns:
        imageUpsampleReal - the inverse Fourier transform of imageInit upsampled by the upsampleFactor. Is an ndarray.
    '''
    imageSize = imageInit.shape
    imageUpsample = np.zeros(tuple((i*upsampleFactor for i in imageSize))) + 0j
    imageUpsample[:imageSize[0], :imageSize[1]] = imageInit
    # plt.figure(1)
    # plt.imshow(np.real(imageUpsample))
    # plt.show(block = True)
    imageUpsample = np.roll(np.roll(imageUpsample, -int(imageSize[0]/2), 0), -int(imageSize[1]/2),1)
    imageUpsampleReal = np.real(np.fft.ifft2(imageUpsample))
    # plt.figure(1)
    # plt.imshow(imageUpsampleReal)
    # plt.show(block = True)
    return imageUpsampleReal

def dftUpsample(imageCorr, upsampleFactor, xyShift):
    '''
    This performs a matrix multiply DFT around a small neighboring region of the inital correlation peak. By using the matrix multiply DFT to do the Fourier upsampling, the efficiency is wildly improved. This is adapted from the subfuction dftups found in the dftregistration function on the Matlab File Exchange.

    https://www.mathworks.com/matlabcentral/fileexchange/18401-efficient-subpixel-image-registration-by-cross-correlation

    The matrix multiplication DFT is from Manuel Guizar-Sicairos, Samuel T. Thurman, and James R. Fienup, "Efficient subpixel image registration algorithms," Opt. Lett. 33, 156-158 (2008). http://www.sciencedirect.com/science/article/pii/S0045790612000778

    Inputs:
        imageCorr - correlation image between two images in Fourier space. ndarray.
        upsampleFactor - scalar integer of how much to upsample.
        xyShift - single pixel shift between images previously computed. Used to center the matrix multiplication on the correlation peak. Is a two element list.
        
    Returns:
        imageUpsample - upsampled image from region around correlation peak. Is a ndarray (and I think the conjugate of the upsampled peak. Has to do wtih order of operations?)
        
    '''
    imageSize = imageCorr.shape
    pixelRadius = 1.5
    numRow = np.ceil(pixelRadius * upsampleFactor)
    numCol = numRow

    colKern = np.exp(
    (-1j * 2 * np.pi / (imageSize[1] * upsampleFactor))
    * (np.fft.ifftshift( (np.arange(imageSize[1])) )
    - np.floor(imageSize[1]/2))
    * (np.arange(numCol) - xyShift[1])[:, np.newaxis]
    ) # I think this can be written differently without the need for np.newaxis. This might require np.outer to compute the matrix itself instead of just using np.dot.

    rowKern = np.exp(
    (-1j * 2 * np.pi / (imageSize[0] * upsampleFactor))
    * (np.arange(numRow) - xyShift[0])
    * (np.fft.ifftshift(np.arange(imageSize[0]))
    - np.floor(imageSize[0]/2))[:, np.newaxis]
    ) # Comment from above applies.

    imageUpsample = np.real(np.dot(np.dot(rowKern.transpose(), imageCorr), colKern.transpose()))

    return imageUpsample

def imageShifter(G2, xyShift):
    '''
    This function multiplies G2 by a plane wave that has the real space effect of shifting ifft2(G2) by [x, y] pixels.

    Inputs:
        G2 - the Fourier transform of an image. ndarray.
        xyShift - a two element list

    Returns:
        G2shift - Fourier shifted G2. ndarray.
    '''
    imageSize = G2.shape
    qx = makeFourierCoords(imageSize[0], 1) # does this need to be a column vector
    if imageSize[1] == imageSize[0]:
        qy = qx
    else:
        qy = makeFourierCoords(imageSize[1], 1)

    G2shift = np.multiply(G2, np.outer( np.exp(-2j * np.pi * qx * xyShift[0]),  np.exp(-2j * np.pi * qy * xyShift[1])))

    return G2shift

def makeFourierCoords(N, pSize):
    '''
    This function creates Fourier coordinates such that (0,0) is in the center of the array.

    Inputs:
        N - the maximum coordinate in the original frame.
        pSize - the pixel size

    Returns:
        q - a single row array that has transformed 0:N to -N/2:N/2, such that the array sizes are the same.
    '''

    N = float(N)
    if N % 2 == 0:
        q = np.roll(np.arange(-N/2, N/2, dtype = 'float64') / (N * pSize), int(-N/2), axis=0)
    else:
        q = np.roll(np.arange((1-N)/2, (N+1)/2) / (N * pSize), int((1-N)/2), axis=0)
    return q
